from __future__ import annotations

from typing import (
    TYPE_CHECKING, Any, Callable, Iterable, List, Literal, Optional, Tuple,
    Union
)

import numpy as np
import xarray as xr

from emsarray.utils import requires_extra

if TYPE_CHECKING:
    from .conventions import Convention

try:
    import cartopy.crs
    from cartopy.feature import GSHHSFeature
    from cartopy.mpl import gridliner
    from matplotlib import animation, patches
    from matplotlib.artist import Artist
    from matplotlib.axes import Axes
    from matplotlib.collections import PatchCollection
    from matplotlib.figure import Figure
    from shapely.geometry import Polygon
    CAN_PLOT = True
    IMPORT_EXCEPTION = None
except ImportError as exc:
    CAN_PLOT = False
    IMPORT_EXCEPTION = exc


__all___ = ['CAN_PLOT', 'plot_on_figure', 'polygon_to_patch']


_requires_plot = requires_extra(extra='plot', import_error=IMPORT_EXCEPTION)


def add_coast(axes: Axes, **kwargs: Any) -> None:
    """
    Add coastlines to an :class:`~matplotlib.axes.Axes`.
    Some default styles are applied:
    the land polygons are light grey and semi-transparent,
    and the coastlines are opaque dark grey.

    Parameters
    ----------
    axes : :class:`matplotlib.axes.Axes`
        The axes to add the coastline to
    **kwargs
        Passed to :meth:`Axes.add_feature() <matplotlib.axes.Axes.add_feature>`.
    """
    kwargs = {
        'facecolor': (0.7, 0.7, 0.7, 0.5),
        'edgecolor': 'darkgrey',
        'linewidth': 0.5,
        **kwargs,
    }
    coast = GSHHSFeature()
    axes.add_feature(coast,  **kwargs)


def add_gridlines(axes: Axes) -> gridliner.Gridliner:
    gridlines = axes.gridlines(draw_labels=True)
    gridlines.top_labels = False
    gridlines.right_labels = False
    return gridlines


def bounds_to_extent(bounds: Tuple[float, float, float, float]) -> List[float]:
    """
    Convert a Shapely bounds tuple to a matplotlib extents.

    A Shapely bounds is a tuple of (min x, min y, max x, max y),
    while a Matplotlib extent is a list of (min x, max x, min y, max y).

    Example
    -------

    .. code-block:: python

        import cartopy.crs as ccrs
        import matplotlib.pyplot as plt
        from emsarray.plot import bounds_to_extent, polygon_to_patch
        from shapely.geometry import Polygon

        polygon = Polygon([
            (148.30, -40.75), (146.47, -41.19), (144.67, -40.62),
            (146.05, -43.53), (146.87, -43.60), (148.30, -40.75),
        ])
        figure = plt.figure(figsize=(10, 8), dpi=100)
        axes = plt.subplot(projection=ccrs.PlateCarree())
        axes.set_extent(bounds_to_extent(polygon.buffer(0.1).bounds))
        axes.add_patch(polygon_to_patch(polygon))
        figure.show()
    """
    minx, miny, maxx, maxy = bounds
    return [minx, maxx, miny, maxy]


@_requires_plot
def polygon_to_patch(polygon: Polygon, **kwargs: Any) -> patches.Polygon:
    """
    Convert a :class:`shapely.geometry.Polygon <Polygon>` to a
    :class:`matplotlib.patches.Polygon`.
    """
    return patches.Polygon(np.transpose(polygon.exterior.xy), **kwargs)


@_requires_plot
def polygons_to_patch_collection(
    polygons: Iterable[Polygon],
    **kwargs: Any,
) -> PatchCollection:
    """
    Convert a list of Shapely :class:`Polygons <Polygon>`
    to a matplotlib :class:`~matplotlib.collections.PatchCollection`.

    Parameters
    ----------
    polygons : iterable of `Polygon`
        The polygons for the patch collection
    **kwargs : Any
        Keyword arguments to pass to the PatchCollection constructor.

    Returns
    -------
    :class:`matplotlib.collections.PatchCollection`
        The PatchCollection made up of the polygons passed in.
    """
    return PatchCollection(map(polygon_to_patch, polygons), **kwargs)


@_requires_plot
def plot_on_figure(
    figure: Figure,
    convention: Convention,
    *,
    scalar: Optional[xr.DataArray] = None,
    vector: Optional[Tuple[xr.DataArray, xr.DataArray]] = None,
    title: Optional[str] = None,
    projection: Optional[cartopy.crs.Projection] = None,
) -> None:
    """
    Plot a :class:`~xarray.DataArray`
    on a :mod:`matplotlib` :class:`~matplotlib.figure.Figure`.

    Parameters
    ----------
    figure
        The :class:`~matplotlib.figure.Figure` instace to plot on.
    convention
        The :class:`~emsarray.conventions.Convention` instance of the dataset.
        This is used to build the polygons and vector quivers.
    scalar : :class:`xarray.DataArray`, optional
        The data to plot as an :class:`xarray.DataArray`.
        This will be passed to :meth:`.Convention.make_patch_collection`.
    vector : tuple of :class:`numpy.ndarray`, optional
        The *u* and *v* components of a vector field
        as a tuple of :class:`xarray.DataArray`.
        These will be passed to :meth:`.Convention.make_quiver`.
    title : str, optional
        The title of the plot. Optional.
    projection : :class:`~cartopy.crs.Projection`
        The projection to use when plotting.
        Optional, defaults to :class:`~cartopy.crs.PlateCarree`.
        This is different to the coordinate reference system for the data,
        which is defined in :attr:`.Convention.data_crs`.
    """
    if projection is None:
        projection = cartopy.crs.PlateCarree()

    axes = figure.add_subplot(projection=projection)
    axes.set_aspect(aspect='equal', adjustable='datalim')

    if scalar is None and vector is None:
        # Plot the polygon shapes for want of anything else to draw
        patches = convention.make_patch_collection()
        axes.add_collection(patches)
        if title is None:
            title = 'Geometry'

    if scalar is not None:
        # Plot a scalar variable on the polygons using a colour map
        patches = convention.make_patch_collection(
            scalar, cmap='jet', edgecolor='face')
        axes.add_collection(patches)
        units = scalar.attrs.get('units')
        figure.colorbar(patches, ax=axes, location='right', label=units)

    if vector is not None:
        # Plot a vector variable using a quiver
        quiver = convention.make_quiver(axes, *vector)
        axes.add_collection(quiver)

    if title:
        axes.set_title(title)

    add_coast(axes)
    add_gridlines(axes)
    axes.autoscale()


@_requires_plot
def animate_on_figure(
    figure: Figure,
    convention: Convention,
    *,
    coordinate: xr.DataArray,
    scalar: Optional[xr.DataArray] = None,
    vector: Optional[Tuple[xr.DataArray, xr.DataArray]] = None,
    title: Optional[Union[str, Callable[[Any], str]]] = None,
    projection: Optional[cartopy.crs.Projection] = None,
    interval: int = 1000,
    repeat: Union[bool, Literal['cycle', 'bounce']] = True,
) -> animation.FuncAnimation:
    """
    Plot a :class:`xarray.DataArray`
    on a :mod:`matplotlib` :class:`~matplotlib.figure.Figure`
    as a :class:`~matplotlib.animation.FuncAnimation`.

    Parameters
    ----------
    figure : :class:`matplotlib.figure.Figure`
        The :class:`~matplotlib.figure.Figure` instace to plot on.
    convention
        The :class:`~emsarray.conventions.Convention` instance of the dataset.
        This is used to build the polygons and vector quivers.
    coordinate : :class:`xarray.DataArray`
        The coordinate values to vary across frames in the animation.
    scalar : :class:`xarray.DataArray`, optional
        The data to plot as an :class:`xarray.DataArray`.
        This will be passed to :meth:`.Convention.make_patch_collection`.
        It should have horizontal dimensions appropriate for this convention,
        and a dimension matching the ``coordinate`` parameter.
    vector : tuple of :class:`numpy.ndarray`, optional
        The *u* and *v* components of a vector field
        as a tuple of :class:`xarray.DataArray`.
        These will be passed to :meth:`.Convention.make_quiver`.
        These should have horizontal dimensions appropriate for this convention,
        and a dimension matching the ``coordinate`` parameter.
    title : str or callable, optional
        The title for each frame of animation.
        Optional, will default to the coordinate value for each frame.

        If this is a string, :meth:`str.format` will be called
        with the coordinate value for each frame.

        If this is a callable,
        it will be called with the coordinate value for each frame,
        and it should return a string for the title.
    projection : :class:`~cartopy.crs.Projection`
        The projection to use when plotting.
        Optional, defaults to :class:`~cartopy.crs.PlateCarree`.
        This is different to the coordinate reference system for the data,
        which is defined in :attr:`.Convention.data_crs`.
    interval : int
        The interval between frames of animation
    repeat : {True, False, 'cycle', 'bounce'}
        Whether to repeat the animation or not.
        ``True`` and ``'cycle'`` will play the animation on loop forever.
        ``'bounce'`` will play the animation forward, then backward, then repeat.
        ``False`` will play the animation once then stop.

    Returns
    -------
    :class:`matplotlib.animation.Animation`
    """
    if projection is None:
        projection = cartopy.crs.PlateCarree()

    axes = figure.add_subplot(projection=projection)
    axes.set_aspect(aspect='equal', adjustable='datalim')
    axes.title.set_animated(True)

    patches = None
    if scalar is not None:
        # Plot a scalar variable on the polygons using a colour map
        scalar_values = convention.make_linear(scalar).values[:, convention.mask]
        patches = convention.make_patch_collection(
            cmap='jet', edgecolor='face',
            clim=(np.nanmin(scalar_values), np.nanmax(scalar_values)))
        axes.add_collection(patches)
        patches.set_animated(True)
        units = scalar.attrs.get('units')
        figure.colorbar(patches, ax=axes, location='right', label=units)

    quiver = None
    if vector is not None:
        # Plot a vector variable using a quiver
        vector_u_values, vector_v_values = (
            convention.make_linear(vec).values
            for vec in vector)
        # Quivers must start with some data.
        # Vector arrows are scaled using this initial data.
        # Find the absolute maximum value in all directions for initial data
        # to make the autoscaling behave across all frames.
        coordinate_dim = coordinate.dims[0]
        initial_u, initial_v = (
            abs(vec).max(dim=coordinate_dim, skipna=True)
            for vec in vector)
        quiver = convention.make_quiver(axes, initial_u, initial_v)
        quiver.set_animated(True)
        axes.add_collection(quiver)

    # Draw a coast overlay
    add_coast(axes)
    gridlines = add_gridlines(axes)
    axes.autoscale()

    repeat_arg = True
    coordinate_indexes = np.arange(coordinate.size)
    if repeat is False:
        repeat_arg = False
    elif repeat == 'bounce':
        coordinate_indexes = np.concatenate(
            (coordinate_indexes, coordinate_indexes[1:-1][::-1]))

    # Set up the animation
    coordinate_callable: Callable[[Any], str]
    if title is None:
        coordinate_callable = str
    elif isinstance(title, str):
        coordinate_callable = title.format
    else:
        coordinate_callable = title

    def animate(index: int) -> Iterable[Artist]:
        changes: List[Artist] = []
        coordinate_value = coordinate.values[index]
        axes.title.set_text(coordinate_callable(coordinate_value))
        changes.append(axes.title)
        changes.extend(gridlines.xline_artists)
        changes.extend(gridlines.yline_artists)

        if patches is not None:
            patches.set_array(scalar_values[index])
            changes.append(patches)

        if quiver is not None:
            quiver.set_UVC(vector_u_values[index], vector_v_values[index])
            changes.append(quiver)

        return changes

    # Draw the figure to force everything to compute its size,
    # for vectors to be initializes, etc.
    figure.draw_without_rendering()

    # Set the first frame of data
    animate(0)

    # Make the animation
    return animation.FuncAnimation(
        figure, animate, frames=coordinate_indexes,
        interval=interval, repeat=repeat_arg, blit=True,
        init_func=lambda: animate(0))



"""
Utilities to make writing robust and friendly command line scripts easier.
"""
import argparse
import contextlib
import decimal
import json
import logging.config
import re
import sys
import textwrap
from functools import wraps
from pathlib import Path
from typing import Callable, Iterator, List, Optional, Protocol

from shapely.geometry import box, shape
from shapely.geometry.base import BaseGeometry

from .exceptions import CommandException

cli_logger = logging.getLogger('emsarray.cli')
error_logger = cli_logger.getChild('errors')
uncaught_exception_logger = error_logger.getChild('uncaught')
command_exception_logger = error_logger.getChild('command')


class MainCallable(Protocol):
    def __call__(
        self,
        argv: Optional[List[str]] = None,
        handle_errors: bool = True,
    ) -> None:
        ...


class NamespaceCallable(Protocol):
    def __call__(self, options: argparse.Namespace) -> None:
        ...


def console_entrypoint(
    add_arguments: Callable[[argparse.ArgumentParser], None]
) -> Callable[[NamespaceCallable], MainCallable]:
    """
    Decorator that turns a function in to a console entrypoint, suitable to use
    as a ``main()`` function. It will automatically set up loggers, make an
    argument parser, add verbosity flags, and handle unexpected errors.

    :func:`.console_entrypoint` takes one argument, a function that sets up an
    :class:`argparse.ArgumentParser` instance.
    :func:`.console_entrypoint` will use this to configure an :class:`~argparse.ArgumentParser`,
    parse the command line flags, and pass the options to the decorated function.

    Example
    -------

    This is a complete example of a command line program that will clip an
    input file using a geometry.

    .. code-block:: python

        #!/usr/bin/env python3

        import emsarray
        import tempfile
        from emsarray.cli.utils import console_entrypoint, geometry_argument
        from pathlib import Path

        def command_line_flags(parser: argparse.ArgumentParser) -> None:
            parser.add_argument('input_file', type=Path)
            parser.add_argument('clip', type=geometry_argument)
            parser.add_argument('output_file', type=Path)

        @console_entrypoint(command_line_flags)
        def main(options: argparse.Namespace) -> None:
            dataset = emsarray.open_dataset(options.input_file)
            with tempfile.TemporaryDirectory() as temp_dir:
                dataset.clip(options.clip, work_dir=Path(temp_dir))
                dataset.ems.to_netcdf(options.output_file)

        if __name__ == '__main__':
            main()

    This example is more-or-less exactly what
    the :func:`.console_entrypoint` decorator does,
    where the decorated function provides the rest of the implementation.

    .. code-block:: python

        @nice_console_errors()
        def main(argv: Optional[List[str]]) -> None:
            parser = argparse.ArgumentParser()
            add_verbosity_group(parser)
            command_line_flags(parser)
            options = parser.parse_args(argv)
            set_verbosity(options.verbosity)

            ...  # Continue on with the rest of the program

    See also
    --------

    :func:`.nice_console_errors`
    :func:`.set_verbosity`
    :func:`.add_verbosity_group`
    """
    def decorator(
        fn: NamespaceCallable,
    ) -> MainCallable:
        @wraps(fn)
        def wrapper(
            argv: Optional[List[str]] = None,
            handle_errors: bool = True,
        ) -> None:
            parser = argparse.ArgumentParser(
                formatter_class=DoubleNewlineDescriptionFormatter,
                add_help=False)
            logging_group = parser.add_argument_group('logging options')
            logging_group.add_argument(
                '-h', '--help', action='help',
                help="Show this help message and quit")
            add_verbosity_group(logging_group)
            add_arguments(parser)
            options = parser.parse_args(argv)
            set_verbosity(options.verbosity)
            cli_logger.debug("Command line options:")
            for name, value in sorted(vars(options).items(), key=lambda i: i[0]):
                cli_logger.debug("%s = %r", name, value)

            if handle_errors:
                with nice_console_errors():
                    fn(options)
            else:
                fn(options)

        return wrapper
    return decorator


@contextlib.contextmanager
def nice_console_errors() -> Iterator:
    """
    A decorator or context manager that sends uncaught errors to the
    appropriate loggers and then quits.
    """
    try:
        yield
    except OSError as err:
        os_error_logger = command_exception_logger.getChild(type(err).__name__)
        os_error_logger.exception(str(err))
        sys.exit(2)
    except CommandException as err:
        # This is a deliberate error raised to exit the program early.
        # Tracebacks are suppressed.
        command_exception_logger.error(err.message)
        sys.exit(err.code)
    except Exception as err:
        uncaught_exception_logger.exception('Uncaught exception: ' + str(err))
        sys.exit(3)
    except KeyboardInterrupt:
        # Handling KeyboardInterrupt properly is tricky, and full of opinions.
        # This is close enough. It will exit with a non-zero exit code,
        # will allow 'atexit' handlers to run and clean things up,
        # but will not set WEXITSTATUS
        cli_logger.info('Caught KeyboardInterrupt, exiting')
        print("")
        sys.exit(1)


class DoubleNewlineDescriptionFormatter(argparse.HelpFormatter):
    def _fill_text(self, text: str, width: int, indent: str) -> str:
        fill_text = super(DoubleNewlineDescriptionFormatter, self)._fill_text

        return '\n\n'.join(
            fill_text(paragraph, width, indent)
            if not paragraph.startswith(' ')
            else paragraph
            for paragraph in textwrap.dedent(text).split('\n\n')
        )


def add_verbosity_group(parser: argparse._ActionsContainer) -> None:
    """
    Add ``--verbose`` and ``--silent`` mutually exclusive flags to an
    :class:`~argparse.ArgumentParser`.
    """
    parser.set_defaults(verbosity=1)
    verbosity = parser.add_mutually_exclusive_group()
    verbosity.add_argument(
        '-v', '--verbose', action='count', dest='verbosity',
        help=(
            "Print more progress information to the console. "
            "Repeating this flag increases the verbosity further."
        ))
    verbosity.add_argument(
        '-q', '--silent', action='store_const', const=0, dest='verbosity',
        help="Print nothing except errors to the console.")


def set_verbosity(level: int) -> None:
    """
    Configure a sensible logging set up, with configurable verbosity.
    """
    if level <= 0:
        level_str = 'ERROR'
    if level == 1:
        level_str = 'WARNING'
    elif level == 2:
        level_str = 'INFO'
    elif level >= 3:
        level_str = 'DEBUG'

    logging.captureWarnings(True)
    logging.config.dictConfig({
        'version': 1,
        'formatters': {
            'default': {
                'format': '[%(asctime)s %(levelname)s] %(name)s: %(message)s',
                'datefmt': '%Y-%m-%d %H:%M:%S',
            },
            'error': {
                'format': '%(message)s',
            }
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'formatter': 'default',
                'level': level_str,
            },
            'error': {
                'class': 'logging.StreamHandler',
                'formatter': 'error',
                'level': level_str,
            },
        },
        'loggers': {
            'emsarray': {'handlers': ['console'], 'level': level_str},
            'emsarray.cli.errors': {
                'handlers': ['error'], 'level': level_str, 'propagate': False
            },
            'py.warnings': {'handlers': ['console'], 'level': 'WARNING'},
            '__main__': {'handlers': ['console'], 'level': level_str},
            'shapely': {'handlers': [], 'level': 60},
        },
        'root_logger': {'handlers': ['console'], 'level': 'WARNING'}
    })


NUMBER = r'\d+(?:_\d+)*'
DECIMAL = rf'({NUMBER}|{NUMBER}\.|\.{NUMBER}|{NUMBER}\.{NUMBER})'
bounds_re = re.compile(r'\s*,\s*'.join([DECIMAL] * 4))


def geometry_argument(argument_string: str) -> BaseGeometry:
    """Try and make some geometry from an argument.
    The following things are tried in order:

    If the argument consists of four comma separated numbers,
    this is converted to a bounding box.
    The numbers are interpreted as (lon min, lat min, lon max, lat max).

    If the argument can be parsed as JSON, it is assumed to be geojson and
    passed to :func:`shapely.geometry.shape`.

    Finally, the argument is interpreted as a file.
    The file extension is used to guess the type of file.
    Currently geojson files are the only supported type

    Example
    -------

    .. code-block:: python

        >>> parser = argparse.ArgumentParser()
        >>> parser.add_argument('bounds', type=geometry_argument)
        >>> options = parser.parse_args(['1,2,3,4'])
        >>> print(options.bounds)
        POLYGON (( 3 4, 1 4, 1 2, 3 2, 3 4 ))
        >>> options = parser.parse_args(['{"type": "Point", "coordinates": [1, 2]}'])
        >>> print(options.bounds)
        POINT (1 2)
        >>> options = parser.parse_args(['./path/to/polygon.geojson'])
        >>> print(options.bounds)
        POLYGON (( 0 1, 1 0, 2 1, 1 2, 0 1 ))
    """
    bounds_match = bounds_re.match(argument_string)
    if bounds_match is not None:
        try:
            return box(*map(float, bounds_match.groups()))
        except ValueError:
            pass

    try:
        json_value = json.loads(argument_string)
    except (ValueError, KeyError):
        pass
    else:
        try:
            return shape(json_value)
        except Exception as err:
            # `shape` raises lots of errors - IndexError, AttributeError,
            # KeyError. Being more specific than Exception is hard.
            raise argparse.ArgumentTypeError(f"Invalid geojson string: {err}") from err

    geometry_path = Path(argument_string)
    if not geometry_path.exists():
        raise argparse.ArgumentTypeError(
            "Argument can not be parsed as bounds, geojson, "
            "or a path to an existing geofile")

    if geometry_path.suffix in {'.geojson', '.json'}:
        try:
            with geometry_path.open('r') as f:
                return shape(json.load(f))
        except Exception as err:
            raise argparse.ArgumentTypeError(f"File is not valid geojson: {err}") from err

    else:
        raise argparse.ArgumentTypeError("Unsupported file type")


def bounds_argument(bounds_string: str) -> BaseGeometry:
    """
    Parse a comma separated string of (lon_min, lat_min, lon_max, lat_max)
    in to a :class:`shapely.geometry.Polygon`. Used as an :mod:`argparse`
    parameter ``type``.

    Example
    -------

    .. code-block:: python

        >>> parser = argparse.ArgumentParser()
        >>> parser.add_argument('bounds', type=bounds_argument)
        >>> options = parser.parse_args(['1,2,3,4'])
        >>> print(options.bounds)
        POLYGON (( 3 4, 1 4, 1 2, 3 2, 3 4 ))
    """
    match = bounds_re.match(bounds_string)
    if match is not None:
        try:
            return box(*map(float, match.groups()))
        except decimal.DecimalException:
            pass
    raise argparse.ArgumentTypeError("Expecting four comma separated numbers for bounds")



import warnings
from typing import Generic, Iterable, Tuple, TypeVar, Union, cast

import numpy as np
import shapely
from packaging.version import parse
from shapely.errors import ShapelyDeprecationWarning
from shapely.geometry.base import BaseGeometry
from shapely.strtree import STRtree

shapely_version = parse(shapely.__version__)
v2 = parse("2")
v1_8_3 = parse("1.8.3")

T = TypeVar("T")


class SpatialIndex(Generic[T]):
    """
    A wrapper around a shapely STRtree that can associate metadata with
    geometries.

    This also handles the version differences in STRtree between
    shapely ~= 1.8.x and shapely >= 2.0.0
    """
    items: np.ndarray
    index: STRtree

    dtype: np.dtype = np.dtype([('geom', np.object_), ('data', np.object_)])

    def __init__(
        self,
        items: Union[np.ndarray, Iterable[Tuple[BaseGeometry, T]]],
    ):
        self.items = np.array(items, dtype=self.dtype)

        if shapely_version >= v2:
            self.index = STRtree(self.items['geom'])
        elif shapely_version >= v1_8_3:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", ShapelyDeprecationWarning)
                self.index = STRtree(geoms=self.items['geom'])
        else:
            self.index = STRtree(geoms=self.items['geom'])

    def query(
        self,
        geom: BaseGeometry,
    ) -> np.ndarray:
        if shapely_version >= v2:
            indices = self.index.query(geom)
        else:
            indices = self.index._query(geom)
        return cast(np.ndarray, self.items.take(indices))

    def nearest(
        self,
        geom: BaseGeometry,
    ) -> np.ndarray:
        if shapely_version >= v2:
            indices = self.index.nearest(geom)
        else:
            indices = self.index._nearest(geom)
        return cast(np.ndarray, self.items.take(indices))



"""Utility functions for working with datasets.
These are low-level functions that apply fixes to datasets,
or provide functionality missing in xarray.
Most users will not have to call these functions directly.

See also
--------
:mod:`emsarray.operations`
"""
from __future__ import annotations

import datetime
import functools
import itertools
import logging
import textwrap
import time
from types import TracebackType
from typing import (
    Any, Callable, Hashable, Iterable, List, Literal, Mapping, MutableMapping,
    Optional, Tuple, Type, TypeVar, Union, cast
)

import cftime
import netCDF4
import numpy as np
import pandas as pd
import pytz
import shapely
import xarray as xr
from packaging.version import Version
from xarray.coding import times
from xarray.core.common import contains_cftime_datetimes

from emsarray.types import Pathish

logger = logging.getLogger(__name__)

DEFAULT_CALENDAR = 'proleptic_gregorian'


_T = TypeVar("_T")
_Exception = TypeVar("_Exception", bound=BaseException)


class PerfTimer:
    __slots__ = ('_start', '_stop', 'running')

    _start: float
    _stop: float
    running: bool

    def __init__(self) -> None:
        self.running = False

    def __enter__(self) -> PerfTimer:
        if self.running:
            raise RuntimeError("Timer is already running")
        self.running = True
        self._start = time.perf_counter()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[_Exception]],
        exc_value: Optional[_Exception],
        traceback: TracebackType
    ) -> Optional[bool]:
        self._stop = time.perf_counter()
        self.running = False
        return None

    @property
    def elapsed(self) -> float:
        if self.running:
            raise RuntimeError("Timer is currently running")
        if not hasattr(self, "_start"):
            raise RuntimeError("Timer has not started yet")
        return self._stop - self._start


def to_netcdf_with_fixes(
    dataset: xr.Dataset,
    path: Pathish,
    time_variable: Optional[Hashable] = None,
    **kwargs: Any,
) -> None:
    """Saves a :class:`xarray.Dataset` to a netCDF4 file,
    applies various fixes to make it compatible with CSIRO software.

    Specifically, this:

    * prevents superfluous ``_FillValue`` attributes being added
      using :func:`.utils.disable_default_fill_value`,
    * Reformats time units that :mod:`xarray` struggles with
      using :func:`.utils.fix_bad_time_units`,
    * Reformats time units after saving to make it compatible with EMS
      using :func:`.utils.fix_time_units_for_ems`

    Parameters
    ----------
    dataset
        The :class:`xarray.Dataset` to save
    path
        Where to save the dataset
    time_variable
        The name of the time variable which needs fixing.
        Optional, if not provided the time variable will not be fixed for EMS.
    kwargs
        Any extra kwargs are passed to :meth:`xarray.Dataset.to_netcdf()`
    """
    # Make a shallow copy of the dataset so we can modify it before saving.
    # A shallow copy only copies the structure and attributes, but leaves
    # the data alone.
    dataset = dataset.copy(deep=False)

    # Fix default xarray behaviour around automatic _FillValues
    disable_default_fill_value(dataset)

    # Some time units are formatted in a way that will cause xarray to fail
    # when saving the dataset. This fixes that formatting on the way out.
    fix_bad_time_units(dataset)

    dataset.to_netcdf(path, **kwargs)
    if time_variable is not None:
        fix_time_units_for_ems(path, time_variable)


def format_time_units_for_ems(units: str, calendar: Optional[str] = DEFAULT_CALENDAR) -> str:
    """
    Reformat a given time unit string to an EMS-compatible string. ``xarray``
    will always format time unit strings using ISO8601 strings with
    a ``T`` separator and no space before the timezone. EMS is unable to parse
    this, and needs spaces between the date, time, and timezone components.

    Parameters
    ----------
    units
        A CF 'units' description of a time variable.
    calendar
        A CF 'calendar' attribute. Defaults to `"proleptic_gregorian"`.

    Returns
    -------
    str
        A new CF 'units' string, representing the same time, but formatted for EMS.

    Example
    -------
    >>> format_time_units_for_ems("days since 1990-01-01T00:00:00+10:00")
    "days since 1990-01-01 00:00:00 +10:00"
    """
    period, date_string = cftime._datesplit(units)
    time_bits = cftime._parse_date(date_string.strip())
    offset_total = time_bits[-1]
    tzinfo = pytz.FixedOffset(offset_total)

    reference_datetime = cftime.num2pydate(0, units, calendar)
    # datetimes come out of num2pydate naive and in UTC
    # This will put them in the correct timezone
    offset_datetime = reference_datetime.replace(tzinfo=pytz.UTC).astimezone(tzinfo)

    offset_hours, offset_minutes = divmod(int(time_bits[-1]), 60)
    offset_string = f'{offset_hours:+d}:{offset_minutes:02d}'

    new_units = f'{period} since {offset_datetime:%Y-%m-%d %H:%M:%S} {offset_string}'

    # Do a quick check that the reference time comes out the same when parsed
    # by cftime. As we are not adjusting the time data itself, the reference
    # time must not change.
    if cftime.num2pydate(0, new_units, calendar) != reference_datetime:
        raise ValueError(
            "New units does not resolve to the same reference time! "
            f"Existing: {units!r}, new: {new_units!r}"
        )

    return new_units


def fix_time_units_for_ems(
    dataset_path: Pathish,
    variable_name: Hashable,
) -> None:
    """
    Updates time units in a file so they are compatible with EMS.
    EMS only supports parsing a subset of valid time unit strings.

    When saving :class:`xarray.Dataset` objects, any time-based variables will be
    saved with a ``units`` like ``"days since 1990-01-01T00:00:00+10:00"`` - a full
    ISO 8601 date string. EMS is old and grumpy, and only accepts time units
    with a format like ``"days since 1990-01-01 00:00:00 +10"``.

    This function will do an in-place update of the time variable units in
    a dataset on disk. It will not recalculate any values, merely update the
    attribute.

    Parameters
    ----------
    dataset_path
        The path to the dataset on disk.
    variable_name
        The name of the time variable in the dataset to fix.
    """

    with netCDF4.Dataset(dataset_path, 'r+') as dataset:
        variable = dataset.variables[variable_name]

        units = cast(str, variable.getncattr('units'))
        calendar = cast(str, variable.getncattr('calendar') or DEFAULT_CALENDAR)

        variable.setncattr('units', format_time_units_for_ems(units, calendar))

        dataset.sync()


def _get_variables(dataset_or_array: Union[xr.Dataset, xr.DataArray]) -> List[xr.Variable]:
    if isinstance(dataset_or_array, xr.Dataset):
        return list(dataset_or_array.variables.values())
    else:
        return [dataset_or_array.variable]


def disable_default_fill_value(dataset_or_array: Union[xr.Dataset, xr.DataArray]) -> None:
    """
    Update all variables on this dataset or data array and disable the
    automatic ``_FillValue`` :mod:`xarray` sets. An automatic fill value can spoil
    ``missing_value``, violate CF conventions for coordinates, and generally
    change a dataset that was loaded from disk in unintentional ways.

    Parameters
    ----------
    dataset_or_array
        The :class:`xarray.Dataset` or :class:`xarray.DataArray` to update
    """
    for variable in _get_variables(dataset_or_array):
        if (
            issubclass(variable.dtype.type, np.floating)
            and "_FillValue" not in variable.encoding
            and "_FillValue" not in variable.attrs
        ):
            variable.encoding["_FillValue"] = None


def fix_bad_time_units(dataset_or_array: Union[xr.Dataset, xr.DataArray]) -> None:
    """Some datasets have a time units string that causes xarray to raise an
    error when saving the dataset. The unit string is parsed when reading the
    dataset just fine, only saving is the issue. This function will check for
    these bad unit strings and change them in place to something xarray can
    handle.

    This issue was fixed in https://github.com/pydata/xarray/pull/6049
    and released in version 0.21.0.
    Once the minimum supported version of xarray is 0.21.0 or higher
    this entire function can be removed.
    """
    if Version(xr.__version__) >= Version('0.21.0'):
        return

    for variable in _get_variables(dataset_or_array):
        # This is the same check xarray uses in xarray.coding.times.CFDatetimeCoder
        is_datetime = (
            np.issubdtype(variable.data.dtype, np.datetime64)
            or contains_cftime_datetimes(variable)
        )
        if is_datetime and 'units' in variable.encoding:
            if 'units' not in variable.encoding:
                # xarray.open_mfdataset() does not propagate the `encoding` dict.
                # This is where the 'units' and 'calendar' attributes are stored.
                # We can't do anything here without those attributes.
                continue
            units = variable.encoding['units']
            calendar = variable.encoding.get('calendar')
            try:
                delta, timestamp = times._unpack_netcdf_time_units(units)
            except ValueError:
                continue  # Don't bother fixing this one - too broken

            try:
                pd.Timestamp(timestamp)
                continue  # These units are formatted fine and don't need fixing
            except ValueError:
                pass

            # This function uses cftime to parse the datestamp and reformat it
            # to something nice, which should take care of the issue for us.
            variable.encoding['units'] = format_time_units_for_ems(units, calendar)


def dataset_like(sample_dataset: xr.Dataset, new_dataset: xr.Dataset) -> xr.Dataset:
    """
    Take an example dataset, and another dataset with identical variable names
    and coordinates, and rearrange the new dataset to have identical ordering
    to the sample. Useful for making a multi-file dataset resemble a sample
    dataset, for example fter masking and saving each variable one-by-one to
    a file.

    Parameters
    ----------
    sample_dataset
        The :class:`xarray.Dataset` to copy the order and attributes from
    new_dataset
        The :class:`xarray.Dataset` to copy the data from.

    Returns
    -------
    :class:`xarray.Dataset`
        A new dataset with attributes and orderings taken from
        ``sample_dataset`` and data taken from ``new_dataset``.
    """
    like_dataset = xr.Dataset(
        # The keys are listed in-order in sample_dataset.
        # Remaking `new_dataset` using this order will fix most things
        data_vars={key: new_dataset[key] for key in sample_dataset.data_vars.keys()},
        coords={key: new_dataset[key] for key in sample_dataset.coords.keys()},
    )
    like_dataset.attrs = new_dataset.attrs.copy()
    like_dataset.encoding = new_dataset.encoding.copy()

    # Copy over attributes and encodings
    _update_no_clobber(sample_dataset.attrs, like_dataset.attrs)
    _update_no_clobber(sample_dataset.encoding, like_dataset.encoding)
    for key, sample_variable in sample_dataset.variables.items():
        new_variable = like_dataset.variables[key]
        _update_no_clobber(sample_variable.attrs, new_variable.attrs)
        _update_no_clobber(sample_variable.encoding, new_variable.encoding)

    # Done!
    return like_dataset


def _update_no_clobber(source: Mapping[Hashable, Any], dest: MutableMapping[Hashable, Any]) -> None:
    """
    Update `dest` dict with values from `source`, without clobbering existing
    keys in `dest`.
    """
    for key, value in source.items():
        if key not in dest:
            dest[key] = value


def extract_vars(
    dataset: xr.Dataset,
    variables: Iterable[Hashable],
    keep_bounds: bool = True,
    errors: Literal['raise', 'ignore'] = 'raise',
) -> xr.Dataset:
    """Extract a set of variables from a dataset, dropping all others.

    This is approximately the opposite of :meth:`xarray.Dataset.drop_vars`.

    Parameters
    ----------
    dataset
        The dataset to extract the variables from
    variables
        A list of variable names
    keep_bounds
        If true (the default), additionally keep any `bounds` variables for the
        included variables and all coordinates.
    errors : {"raise", "ignore"}, optional
        If 'raise' (default), raises a :exc:`ValueError` error if any of the
        variable passed are not in the dataset. If 'ignore', any given names
        that are in the dataset are kept and no error is raised.

    Returns
    -------
    :class:`xr.Dataset`
        A new dataset with only the named variables included.

    See also
    --------
    :meth:`xarray.Dataset.drop_vars`
    """
    variables = set(variables)

    if errors == 'raise':
        missing_variables = variables - set(dataset.variables.keys())
        if missing_variables:
            raise ValueError(
                f"Variables {sorted(missing_variables, key=str)!r} "
                "do not exist in the dataset"
            )

    if keep_bounds:
        variables_and_coordinates = variables | set(dataset.coords.keys())
        variables = variables | {
            dataset[var].attrs['bounds'] for var in variables_and_coordinates
            if 'bounds' in dataset[var].attrs
        }

    drop_vars = list(set(dataset.data_vars.keys()) - variables)
    drop_vars = [name for name in dataset.data_vars.keys() if name not in variables]
    logger.debug("Dropping variables %r", drop_vars)
    return dataset.drop_vars(drop_vars)


def pairwise(iterable: Iterable[_T]) -> Iterable[Tuple[_T, _T]]:
    """
    Iterate over values in an iterator in pairs.

    Example
    -------
    >>> for a, b in pairwise("ABCD"):
    ...     print(a, b)
    A B
    B C
    C D
    """
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


def dimensions_from_coords(
    dataset: xr.Dataset,
    coordinate_names: List[Hashable],
) -> List[Hashable]:
    """
    Get the names of the dimensions for a set of coordinates.

    Parameters
    ----------
    dataset
        The dataset to get the dimensions from
    coordinate_names
        The names of some coordinate variables.

    Returns
    -------
    list of Hashable
        The name of the relevant dimension for each coordinate variable.
    """
    dimensions = []
    for coordinate_name in coordinate_names:
        # Don't use dataset.coords[], auxilliary coordinates do not appear there
        data_array = dataset.variables[coordinate_name]
        if len(data_array.dims) > 1:
            raise ValueError(
                f"Coordinate variable {coordinate_name} has more "
                "than one dimension: {data_array.dims}")
        dimensions.append(data_array.dims[0])

    return dimensions


def check_data_array_dimensions_match(dataset: xr.Dataset, data_array: xr.DataArray) -> None:
    """
    Check that the dimensions of a :class:`xarray.DataArray`
    match the dimensions of a :class:`xarray.Dataset`.
    This is useful when using the metadata of a particular dataset to display a data array,
    without requiring the data array to be taken directly from the dataset.

    If the dimensions do not match, a ValueError is raised, indicating the mismatched dimension.

    Parameters
    ----------
    dataset
        The dataset used as a reference
    data_array
        The data array to check the dimensions of

    Raises
    ------
    ValueError
        Raised if the dimensions do not match
    """
    for dimension, data_array_size in zip(data_array.dims, data_array.shape):
        if dimension not in dataset.dims:
            raise ValueError(
                f"Data array has unknown dimension {dimension} of size {data_array_size}"
            )

        dataset_size = dataset.dims[dimension]
        if data_array_size != dataset_size:
            raise ValueError(
                "Dimension mismatch between dataset and data array: "
                f"Dataset dimension {dimension} has size {dataset_size}, "
                f"data array has size {data_array_size}"
            )


def move_dimensions_to_end(
    data_array: xr.DataArray,
    dimensions: List[Hashable],
) -> xr.DataArray:
    """
    Transpose the dimensions of a :class:`xarray.DataArray`
    such that the given dimensions appear as the last dimensions,
    in the order given.

    Other dimensions appear as the first dimensions,
    in the same order they are present in the original dataset

    Parameters
    ----------
    `data_array` : :class:`xarray.DataArray`
        The data array to transpose
    `dimensions` : list of Hashable
        The dimensions to move to the end

    Examples
    --------

    .. code-block:: python

        >>> data_array.dims
        ('a', 'b', 'c', 'd')
        >>> transposed = move_dimensions_to_end(data_array, ['c', 'b'])
        >>> transposed.dims
        ('a', 'd', 'c', 'b')
    """
    current_dims = set(data_array.dims)
    if not current_dims.issuperset(dimensions):
        missing = sorted(set(dimensions) - set(current_dims), key=str)
        raise ValueError(f"DataArray does not contain dimensions {missing!r}")

    new_order = [dim for dim in data_array.dims if dim not in dimensions] + dimensions
    if new_order == list(data_array.dims):
        # Don't bother transposing if the dimensions are already correct.
        return data_array.copy(deep=False)
    else:
        return data_array.transpose(*new_order)


def linearise_dimensions(
    data_array: xr.DataArray,
    dimensions: List[Hashable],
    linear_dimension: Optional[Hashable] = None,
) -> xr.DataArray:
    """
    Flatten the given dimensions of a :class:`~xarray.DataArray`.
    Other dimensions are kept as-is.
    This is useful for turning a DataArray with dimensions ('t', 'z', 'y', 'x')
    in to ('t', 'z', 'index').

    Parameters
    ----------

    `data_array` : :class:`xarray.DataArray`
        The data array to linearize
    `dimensions` : list of Hashable
        The dimensions to linearize, in the desired order.
        These dimensions can be in any order and any position in the input data array.
    `linear_dimension` : Hashable, optional
        The name of the new dimension of flattened data.
        Defaults to `index`, or `index_0`, `index_1`, etc if not given.

    Returns
    -------
    :class:`xarray.DataArray`
        A new data array with the specified dimensions flattened.
        Only data, coordinates, and dimensions are set,
        attributes and encodings are not copied over.

    Examples
    --------

    .. code-block:: python

        >>> data_array = xr.DataArray(
        ...     data=np.random.random((3, 5, 7)),
        ...     dims=['x', 'y', 'z'],
        ... )
        >>> flattened = linearise_dimensions(data_array, ['y', 'x'])
        >>> flattened.dims
        ('z', 'index')
        >>> flattened.shape
        (7, 15)
        >>> expected = np.transpose(data_array.isel(z=0).values).ravel()
        >>> all(flattened.isel(z=0).values == expected)
        True
    """
    data_array = move_dimensions_to_end(data_array, dimensions)
    new_shape = data_array.shape[:-len(dimensions)] + (-1,)
    new_data = data_array.values.reshape(new_shape)
    existing_dims = data_array.dims[:-len(dimensions)]

    if linear_dimension is None:
        suffix = 0
        linear_dimension = 'index'
        while linear_dimension in existing_dims:
            linear_dimension = f'index_{suffix}'
            suffix += 1
    new_dims = existing_dims + (linear_dimension,)

    coords = {
        name: coord for name, coord in data_array.coords.items()
        if set(coord.dims).issubset(existing_dims)}
    return xr.DataArray(data=new_data, dims=new_dims, coords=coords)


def datetime_from_np_time(np_time: np.datetime64) -> datetime.datetime:
    """
    Convert a numpy :class:`~numpy.datetime64`
    to a python :class:`~datetime.datetime`.
    Useful when formatting dates for human consumption,
    as a numpy :class:`~numpy.datetime64`
    has no equivalent of :meth:`datetime.datetime.strftime`.

    This does present the possibility of losing precision,
    as a numpy datetime64 has variable accuracy up to an attosecond,
    while Python datetimes have fixed microsecond accuracy.
    A conversion that truncates data is not reported as an error.
    If you're using numpy datetime64 with attosecond accuracy,
    the Python datetime formatting methods are insufficient for your needs anyway.
    """
    return datetime.datetime.fromtimestamp(np_time.item() / 10**9)


class RequiresExtraException(Exception):
    """
    Raised when the optional dependencies for some functionality have not been installed,
    and a function requiring them is called.

    See also
    --------
    :func:`requires_extra`
    """

    message_template = """
        emsarray must be installed with the '{extra}' extra in order to use this functionality:

            $ pip install emsarray[{extra}]

        Or, if installed using conda:

            $ conda install emsarray
    """
    message_template = textwrap.dedent(message_template).strip("\n")

    def __init__(self, extra: str) -> None:
        super().__init__(self.message_template.format(extra=extra))
        self.extra = extra


def requires_extra(
    extra: str,
    import_error: Optional[ImportError],
    exception_class: Type[RequiresExtraException] = RequiresExtraException,
) -> Callable[[_T], _T]:
    if import_error is None:
        return lambda fn: fn

    def error_decorator(fn: _T) -> _T:
        @functools.wraps(fn)  # type: ignore
        def error(*args: Any, **kwargs: Any) -> Any:
            raise exception_class(extra) from import_error
        return error  # type: ignore
    return error_decorator


def make_polygons_with_holes(
    points: np.ndarray,
    *,
    out: Optional[np.ndarray] = None,
) -> np.ndarray:
    """
    Make a :class:`numpy.ndarray` of :class:`shapely.Polygon` from an array of (n, m, 2) points.
    ``n`` is the number of polygons, ``m`` is the number of vertices per polygon.
    If any point in a polygon is :data:`np.nan`,
    that Polygon is skipped and will be :class:`None` in the returned array.

    Parameters
    ----------

    points : np.ndarray
        A (n, m, 2) array. Each row represents the m points of a polygon.
    out : np.ndarray, optional
        Optional. An array to fill with polygons.

    Returns
    -------

    np.ndarray
        The polygons in a array of size n.
    """
    if out is None:
        out = np.full(points.shape[0], None, dtype=np.object_)

    complete_row_indices = np.flatnonzero(np.isfinite(points).all(axis=(1, 2)))
    shapely.polygons(
        points[complete_row_indices],
        indices=complete_row_indices,
        out=out)
    return out



"""
Datasets following the CF conventions with gridded datasets.
Both 1D coordinates and 2D coordinates are supported.
"""
from __future__ import annotations

import abc
import enum
import itertools
import warnings
from contextlib import suppress
from functools import cached_property
from typing import (
    Dict, Generic, Hashable, Optional, Tuple, Type, TypeVar, cast
)

import numpy as np
import xarray as xr
from shapely.geometry.base import BaseGeometry

from emsarray import masking, utils
from emsarray.exceptions import ConventionViolationWarning
from emsarray.types import Pathish

from ._base import Convention, Specificity


class CFGridKind(str, enum.Enum):
    """"""
    face = 'face'


#: A two-tuple of ``(y, x)``.
CFGridIndex = Tuple[int, int]


class CFGridTopology(abc.ABC):
    """
    A topology helper that keeps track of the latitude and longitude coordinates
    in a CF grid dataset.
    """
    def __init__(
        self,
        dataset: xr.Dataset,
        longitude: Optional[Hashable] = None,
        latitude: Optional[Hashable] = None,
    ):
        """
        Construct a new :class:`CFGridTopology` instance for a dataset.

        By default this will introspect the dataset
        looking for a longitude and latitude coordinate variable.
        The ``longitude`` and ``latitude`` parameters
        allow you to manually specify the correct coordinate names
        if the automatic detection fails.
        """
        self.dataset = dataset
        if longitude is not None:
            self.longitude_name = longitude
        if latitude is not None:
            self.latitude_name = latitude

    @cached_property
    def latitude_name(self) -> Hashable:
        """
        The name of the latitude coordinate variable.
        Found by looking for a variable with either a
        ``standard_name = "latitude"`` or
        ``units = "degree_north"``
        attribute.
        """
        try:
            return next(
                name
                for name, variable in self.dataset.variables.items()
                if variable.attrs.get('standard_name') == 'latitude'
                or variable.attrs.get('coordinate_type') == 'latitude'
                or variable.attrs.get('units') == 'degree_north'
            )
        except StopIteration:
            raise ValueError("Could not find latitude coordinate")

    @cached_property
    def longitude_name(self) -> Hashable:
        """
        The name of the longitude coordinate variable.
        Found by looking for a variable with either a
        ``standard_name = "longitude"`` or
        ``units = "degree_east"``
        attribute.
        """
        try:
            return next(
                name for name, variable in self.dataset.variables.items()
                if variable.attrs.get('standard_name') == 'longitude'
                or variable.attrs.get('coordinate_type') == 'longitude'
                or variable.attrs.get('units') == 'degree_east'
            )
        except StopIteration:
            raise ValueError("Could not find longitude coordinate")

    @property
    def latitude(self) -> xr.DataArray:
        """The latitude coordinate variable"""
        return self.dataset[self.latitude_name]

    @property
    def longitude(self) -> xr.DataArray:
        """The longitude coordinate variable"""
        return self.dataset[self.longitude_name]

    @property
    @abc.abstractmethod
    def latitude_bounds(self) -> xr.DataArray:
        """
        Bounds for the latitude coordinate variable.
        If there are no bounds defined on the dataset,
        bounds will be generated.
        """
        pass

    @property
    @abc.abstractmethod
    def longitude_bounds(self) -> xr.DataArray:
        """
        Bounds for the longitude coordinate variable.
        If there are no bounds defined on the dataset,
        bounds will be generated.
        """
        pass

    @property
    @abc.abstractmethod
    def y_dimension(self) -> Hashable:
        """
        The name of the 'y' dimension.
        For 1D coordinates, this is the the latitude dimension.
        For 2D coordinates, this is the first dimension of the coordinate variables.
        """
        pass

    @property
    @abc.abstractmethod
    def x_dimension(self) -> Hashable:
        """
        The name of the 'x' dimension.
        For 1D coordinates, this is the the longitude dimension.
        For 2D coordinates, this is the second dimension of the coordinate variables.
        """
        pass

    @cached_property
    def shape(self) -> Tuple[int, int]:
        """The shape of this grid, as a tuple of ``(y, x)``."""
        dims = self.dataset.dims
        return (dims[self.y_dimension], dims[self.x_dimension])

    @cached_property
    def size(self) -> int:
        """
        The scalar size of this grid.

        Equal to ``np.prod(topology.shape)``,
        i.e., the product of the grid dimensions.
        """
        return int(np.prod(self.shape))


Topology = TypeVar('Topology', bound=CFGridTopology)


class CFGrid(Generic[Topology], Convention[CFGridKind, CFGridIndex]):
    """
    A base class for CF grid datasets.
    There are two concrete subclasses: :class:`CFGrid1D` and :class:`CFGrid2D`.
    """

    grid_kinds = frozenset(CFGridKind)
    default_grid_kind = CFGridKind.face
    topology_class: Type[Topology]

    def __init__(
        self,
        dataset: xr.Dataset,
        *,
        latitude: Optional[Hashable] = None,
        longitude: Optional[Hashable] = None,
        topology: Optional[Topology] = None,
    ) -> None:
        """
        Construct a new :class:`CFGrid` instance.

        Parameters
        ----------
        dataset : :class:`xarray.Dataset`
            A :class:`~xarray.Dataset` that follows the CF conventions.
            The grid coordinates must be one-dimensional.
        latitude : Hashable, optional
            The name of the latitude coordinate variable on this dataset. Optional.
            By default the coordinate variables are found by introspecting the dataset.
            You can use this parameter to override this behaviour.
        longitude : Hashable, optional
            The name of the longitude coordinate variable on this dataset. Optional.
            By default the coordinate variables are found by introspecting the dataset.
            You can use this parameter to override this behaviour.
        topology : :class:`CFGridTopology`, optional
            Optional, allows you to override the default topology helper.
        """
        super().__init__(dataset)
        if latitude is not None and longitude is not None:
            if topology is not None:
                raise TypeError(
                    "Can not pass both latitude and longitude arguments, "
                    "and the topology argument")
            topology = self.topology_class(
                self.dataset, latitude=latitude, longitude=longitude)

        if topology is not None:
            self.topology = topology

    @cached_property
    def topology(self) -> Topology:
        """A :class:`CFGridTopology` helper."""
        return self.topology_class(self.dataset)

    def unravel_index(
        self,
        index: int,
        grid_kind: Optional[CFGridKind] = None,
    ) -> CFGridIndex:
        y, x = map(int, np.unravel_index(index, self.topology.shape))
        return (y, x)

    def ravel_index(self, indices: CFGridIndex) -> int:
        return int(np.ravel_multi_index(indices, self.topology.shape))

    def get_grid_kind_and_size(self, data_array: xr.DataArray) -> Tuple[CFGridKind, int]:
        expected = {self.topology.y_dimension, self.topology.x_dimension}
        dims = set(data_array.dims)
        if dims.issuperset(expected):
            return (CFGridKind.face, self.topology.size)

        expected_sorted = sorted(expected, key=str)
        raise ValueError(f"Data array did not have dimensions {expected_sorted!r}")

    def selector_for_index(self, index: CFGridIndex) -> Dict[Hashable, int]:
        y, x = index
        return {self.topology.y_dimension: y, self.topology.x_dimension: x}

    def drop_geometry(self) -> xr.Dataset:
        dataset = self.dataset.drop_vars([
            self.topology.longitude_name,
            self.topology.latitude_name,
        ])
        dataset.attrs.pop('Conventions', None)

        return dataset

    def make_linear(self, data_array: xr.DataArray) -> xr.DataArray:
        surface_dims = [self.topology.y_dimension, self.topology.x_dimension]
        return utils.linearise_dimensions(data_array, surface_dims)

    def make_clip_mask(
        self,
        clip_geometry: BaseGeometry,
        buffer: int = 0,
    ) -> xr.Dataset:
        topology = self.topology

        intersecting_indices = [
            item.linear_index
            for polygon, item in self.spatial_index.query(clip_geometry)
            if polygon.intersects(clip_geometry)]
        mask = np.full(topology.shape, fill_value=False)
        mask.ravel()[intersecting_indices] = True

        if buffer > 0:
            mask = masking.blur_mask(mask, size=buffer)
        dimensions = [topology.y_dimension, topology.x_dimension]

        return xr.Dataset(
            data_vars={
                'cell_mask': xr.DataArray(data=mask, dims=dimensions),
            },
            coords={
                topology.latitude_name: topology.latitude.copy(),
                topology.longitude_name: topology.longitude.copy(),
            },
            attrs={'type': 'CFGrid mask'},
        )

    def apply_clip_mask(self, clip_mask: xr.Dataset, work_dir: Pathish) -> xr.Dataset:
        return masking.mask_grid_dataset(self.dataset, clip_mask, work_dir)


# 1D coordinate grids

class CFGrid1DTopology(CFGridTopology):
    """
    Collects information about the topology of a gridded dataset
    so that you don't have to.

    This grid has one-dimensional coordinates,
    such as ``lat(lat)`` or ``longitude(x)``.
    """

    @cached_property
    def y_dimension(self) -> Hashable:
        """The name of the latitude dimension, taken from the latitude coordinate"""
        return self.latitude.dims[0]

    @cached_property
    def x_dimension(self) -> Hashable:
        """The name of the latitude dimension, taken from the latitude coordinate"""
        return self.longitude.dims[0]

    def _get_or_make_bounds(self, coordinate: xr.DataArray) -> xr.DataArray:
        with suppress(KeyError):
            bounds = self.dataset.data_vars[coordinate.attrs['bounds']]
            if (
                len(bounds.dims) == 2
                and bounds.dims[0] == coordinate.dims[0]
                and self.dataset.dims[bounds.dims[1]] == 2
            ):
                return bounds
            else:
                expected_dims = (coordinate.dims[0], 2)
                warnings.warn(
                    f"Bounds {bounds.name!r} for coordinate {coordinate.name!r} "
                    f"had invalid dimensions. "
                    f"Expected {expected_dims!r}, "
                    f"found {tuple(bounds.dims)!r}",
                    category=ConventionViolationWarning,
                    stacklevel=4)

        values = coordinate.values
        first_gap = values[1] - values[0]
        last_gap = values[-1] - values[-2]
        mid_points = np.concatenate([
            [values[0] - first_gap / 2],
            (values[1:] + values[:-1]) / 2,
            [values[-1] + last_gap / 2]
        ])
        return xr.DataArray(
            np.stack([mid_points[:-1], mid_points[1:]], axis=-1),
            dims=[coordinate.dims[0], 'bounds'],
        )

    @cached_property
    def latitude_bounds(self) -> xr.DataArray:
        """North/south bounds for each latitude point"""
        return self._get_or_make_bounds(self.latitude)

    @cached_property
    def longitude_bounds(self) -> xr.DataArray:
        """East/west bounds for each longitude point"""
        return self._get_or_make_bounds(self.longitude)


class CFGrid1D(CFGrid[CFGrid1DTopology]):
    """A :class:`.Convention` subclass representing datasets on an axis-aligned grid
    that follows the CF metadata conventions
    and has one dimensional coordinates.
    """
    topology_class = CFGrid1DTopology

    @classmethod
    def check_dataset(cls, dataset: xr.Dataset) -> Optional[int]:
        """
        A dataset is a 1D CF grid if it has one dimensional
        latitude and longitude coordinate variables.

        Many other conventions extend the CF conventions,
        so this matches with low specificity.
        """
        topology = cls.topology_class(dataset)
        try:
            latitude = topology.latitude
            longitude = topology.longitude
        except ValueError:
            return None

        # Must have one dimensional coordinates
        if len(latitude.dims) != 1 or len(longitude.dims) != 1:
            return None

        return Specificity.LOW

    @cached_property
    def polygons(self) -> np.ndarray:
        lon_bounds = self.topology.longitude_bounds.values
        lat_bounds = self.topology.latitude_bounds.values

        # Make a bounds array as if this dataset had 2D coordinates.
        # 1D bounds are (min, max).
        # 2D bounds are (j-i, i-1), (j-1, i+1), (j+1, i+1), (j+1, i-1).
        # The bounds values are repeated as required, are given a new dimension,
        # then repeated along that new dimension.
        # They will come out as array with shape (lat, lon, 4)
        y_size, x_size = self.topology.shape
        lon_bounds_2d = np.tile(lon_bounds[np.newaxis, :, [0, 1, 1, 0]], (y_size, 1, 1))
        lat_bounds_2d = np.tile(lat_bounds[:, np.newaxis, [0, 0, 1, 1]], (1, x_size, 1))

        # points is a (topology.size, 4, 2) array of the corners of each cell
        points = np.stack([lon_bounds_2d, lat_bounds_2d], axis=-1).reshape((-1, 4, 2))

        return utils.make_polygons_with_holes(points)

    @cached_property
    def face_centres(self) -> np.ndarray:
        topology = self.topology
        xx, yy = np.meshgrid(topology.longitude.values, topology.latitude.values)
        centres = np.column_stack((xx.flatten(), yy.flatten()))
        return cast(np.ndarray, centres)


# 2D coordinate grids

class CFGrid2DTopology(CFGridTopology):
    """
    Collects information about the topology of a gridded dataset
    so that you don't have to.

    This grid has two-dimensional coordinates,
    such as ``lat(y, x)`` and ``lon(y, x)``
    """
    @property
    def y_dimension(self) -> Hashable:
        """
        The name of the first dimension of the coordinate variables.
        This is nominally called the ``y`` dimension.
        """
        return self.latitude.dims[0]

    @property
    def x_dimension(self) -> Hashable:
        """
        The name of the second dimension of the coordinate variables.
        This is nominally called the ``x`` dimension.
        """
        return self.latitude.dims[1]

    def _get_or_make_bounds(self, coordinate: xr.DataArray) -> xr.DataArray:
        # Use the bounds defined on the coordinate itself, if any.
        with suppress(KeyError):
            bounds = self.dataset[coordinate.attrs['bounds']]
            if (
                len(bounds.dims) == 3
                and bounds.dims[0] == self.y_dimension
                and bounds.dims[1] == self.x_dimension
                and self.dataset.dims[bounds.dims[2]] == 4
            ):
                return cast(xr.DataArray, bounds)
            else:
                expected_dims = (self.y_dimension, self.x_dimension, 2)
                warnings.warn(
                    f"Bounds {bounds.name!r} for coordinate {coordinate.name!r} "
                    f"had invalid dimensions. "
                    f"Expected {expected_dims!r}, "
                    f"found {tuple(bounds.dims)!r}",
                    category=ConventionViolationWarning,
                    stacklevel=4)

        # The dataset has no bounds defined. Let us make our own.
        #
        # Assume the bounds for each cell are contiguous with the neighbouring cells.
        # The corners are the average of the four surrounding cell centres.
        # On the edges where there are fewer 'surrounding' cells, the cell centres are used.
        # Edge and corner cells will be smaller than the surrounding cells because of this.

        # np.nanmean will return nan for an all-nan column.
        # This is the exact behaviour that we want.
        # numpy emits a warning that can not be silenced when this happens,
        # so that warning is temporarily ignored.
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", "Mean of empty slice", category=RuntimeWarning)
            grid = np.nanmean([
                np.pad(coordinate.values, pad, constant_values=np.nan)
                for pad in itertools.product([(1, 0), (0, 1)], [(1, 0), (0, 1)])
            ], axis=0)

        y_size, x_size = self.shape
        bounds = np.array([
            [
                [grid[y, x], grid[y, x + 1], grid[y + 1, x + 1], grid[y + 1, x]]
                for x in range(x_size)
            ]
            for y in range(y_size)
        ])
        # Any cell that has a `nan` in its bounds will be set to all nan
        cells_with_nans = np.isnan(bounds).any(axis=2)
        bounds[cells_with_nans] = np.nan

        return xr.DataArray(
            bounds,
            dims=[self.y_dimension, self.x_dimension, 'bounds'],
        )

    @cached_property
    def longitude_bounds(self) -> xr.DataArray:
        return self._get_or_make_bounds(self.longitude)

    @property
    def latitude_bounds(self) -> xr.DataArray:
        return self._get_or_make_bounds(self.latitude)


class CFGrid2D(CFGrid[CFGrid2DTopology]):
    """A :class:`.Convention` subclass representing datasets on a curvilinear grid
    that follows the CF metadata conventions
    and has two dimensional coordinates.
    """
    topology_class = CFGrid2DTopology

    @classmethod
    def check_dataset(cls, dataset: xr.Dataset) -> Optional[int]:
        """
        A dataset is a 2D CF grid if it has two dimensional
        latitude and longitude coordinate variables.

        Many other conventions extend the CF conventions,
        so this matches with low specificity.
        """
        topology = cls.topology_class(dataset)
        try:
            latitude = topology.latitude
            longitude = topology.longitude
        except ValueError:
            return None

        # Must have two dimensional coordinates
        if len(latitude.dims) != 2 or len(longitude.dims) != 2:
            return None

        return Specificity.LOW

    @cached_property
    def polygons(self) -> np.ndarray:
        # Construct polygons from the bounds of the cells
        lon_bounds = self.topology.longitude_bounds.values
        lat_bounds = self.topology.latitude_bounds.values

        # points is a (topology.size, 4, 2) array of the corners of each cell
        points = np.stack([lon_bounds, lat_bounds], axis=-1).reshape((-1, 4, 2))

        return utils.make_polygons_with_holes(points)

    @cached_property
    def face_centres(self) -> np.ndarray:
        centres = np.column_stack((
            self.make_linear(self.topology.longitude).values,
            self.make_linear(self.topology.latitude).values,
        ))
        return cast(np.ndarray, centres)



from __future__ import annotations

import logging

import xarray as xr

from .conventions import Convention, get_dataset_convention
from .state import State

logger = logging.getLogger(__name__)


@xr.register_dataset_accessor("ems")
def ems_accessor(dataset: xr.Dataset) -> Convention:
    """Provides the ``.ems`` attribute on xarray Datasets.
    This will make a :class:`~emsarray.conventions.Convention` instance for the dataset,
    using the correct :class:`~emsarray.conventions.Convention` subclass depending on the file type.

    Returns
    -------
    :class:`~emsarray.conventions.Convention`
    """
    state = State.get(dataset)
    if state.convention is not None:
        return state.convention

    convention_class = get_dataset_convention(dataset)
    if convention_class is None:
        raise RuntimeError("Could not determine convention of dataset")

    convention = convention_class(dataset)
    convention.bind()
    return convention


xr.register_dataset_accessor(State.accessor_name)(State)



"""
These operations function on datasets with a depth axis,
such as the output from ocean models.
"""
import warnings
from collections import defaultdict
from typing import Dict, FrozenSet, Hashable, List, Optional, cast

import numpy as np
import xarray as xr

from emsarray import utils


def ocean_floor(
    dataset: xr.Dataset,
    depth_variables: List[Hashable],
    *,
    non_spatial_variables: Optional[List[Hashable]] = None,
) -> xr.Dataset:
    """Make a new :class:`xarray.Dataset` reduced along the given depth
    coordinates to only contain values along the ocean floor.

    Parameters
    ----------
    dataset
        The dataset to reduce.
    depth_variables
        The names of depth coordinate variables.
        For supported conventions, use :meth:`.Convention.get_all_depth_names()`.
    non_spatial_variables
        Optional.
        A list of the names of any non-spatial coordinate variables, such as time.
        The ocean floor is assumed to be static across non-spatial dimensions.
        For supported conventions, use :meth:`.Convention.get_time_name()`.

    Returns
    -------
    :class:`xarray.Dataset`
        A new dataset with values taken from the deepest data.

    Examples
    --------

    .. code-block:: python

        >>> dataset
        <xarray.Dataset>
        Dimensions:  (z: 5, y: 5, x: 5)
        Coordinates:
            lon      (x) int64 0 -1 -2 -3 -4
            lat      (y) int64 0 1 2 3 4
            depth    (z) float64 4.25 3.25 2.25 1.25 0.25
        Dimensions without coordinates: z, y, x
        Data variables:
            temp     (z, y, x) float64 0.0 nan nan nan nan nan ... 4.0 4.0 4.0 4.0 4.0
        >>> operations.ocean_floor(dataset, ['depth'])
        <xarray.Dataset>
        Dimensions:  (y: 5, x: 5)
        Coordinates:
            lon      (x) int64 0 -1 -2 -3 -4
            lat      (y) int64 0 1 2 3 4
        Dimensions without coordinates: y, x
        Data variables:
            temp     (y, x) float64 0.0 1.0 2.0 3.0 4.0 1.0 ... 4.0 4.0 4.0 4.0 4.0

    This operation is relatively efficient,
    but still involves masking every variable that includes a depth axis.
    Where possible, do any time and space slicing before calling this method,
    and drop any variables you are not interested in.

    .. code-block:: python

        >>> operations.ocean_floor(
        ...     big_dataset['temp'].isel(record=0).to_dataset(),
        ...     depth_variables=big_dataset.ems.get_all_depth_names())
        <xarray.Dataset>
        Dimensions:  (y: 5, x: 5)
        Coordinates:
            lon      (x) int64 0 -1 -2 -3 -4
            lat      (y) int64 0 1 2 3 4
        Dimensions without coordinates: y, x
        Data variables:
            temp     (y, x) float64 0.0 1.0 2.0 3.0 4.0 1.0 ... 4.0 4.0 4.0 4.0 4.0

    See also
    --------
    :meth:`.Convention.ocean_floor`
    :meth:`.Convention.get_all_depth_names`
    :func:`.normalize_depth_variables`
    :func:`.utils.extract_vars`
    """
    # Consider both curvilinear SHOC datasets and UGRID COMPASS datasets.
    #
    # SHOC datasets have four 'grids': faces, left edges, back edges, and vertices.
    # Each of these grids has a two dimensional (j, i) index.
    #
    # COMPASS datasets have up to three 'grids': faces, edges, and nodes.
    # Each of these grids has a one dimensional index.
    #
    # Data variables might be defined on any one of these grids,
    # might have non-spatial (e.g. time) dimensions,
    # and might have a depth axis.
    # We assume that a data variable has at most one depth axis,
    # and that for a combination of depth dimension and spatial dimensions,
    # the ocean floor is static.

    dataset = normalize_depth_variables(
        dataset, depth_variables,
        positive_down=True, deep_to_shallow=False)

    if non_spatial_variables is None:
        non_spatial_variables = []

    # The name of all the relevant _dimensions_, not _coordinates_
    depth_dimensions = utils.dimensions_from_coords(dataset, depth_variables)
    non_spatial_dimensions = utils.dimensions_from_coords(dataset, non_spatial_variables)

    for depth_dimension in sorted(depth_dimensions, key=hash):
        dimension_sets: Dict[FrozenSet[Hashable], List[Hashable]] = defaultdict(list)
        for name, variable in dataset.data_vars.items():
            if depth_dimension not in variable.dims:
                continue  # Skip data variables without this depth dimension

            spatial_dimensions = frozenset(variable.dims).difference(
                {depth_dimension}, non_spatial_dimensions)
            if not spatial_dimensions:
                continue  # Skip data variables with no other spatial dimenions

            dimension_sets[spatial_dimensions].append(name)

        for spatial_dimensions, variable_names in dimension_sets.items():
            # We now have a set of spatial_dimenions,
            # and a list of data variable_names.
            # We assume that a specific combination of depth variable and
            # spatial dimensions has a static ocean floor.
            # We find the ocean floor for one of the data variables,
            # and then use that to mask out each data variable in turn.

            # Get an example data array and drop all the non-spatial dimensions.
            data_array = dataset.data_vars[variable_names[0]].isel(
                {name: 0 for name in non_spatial_dimensions},
                drop=True, missing_dims='ignore')
            # Then find the ocean floor indices.
            ocean_floor_indices = _find_ocean_floor_indices(
                data_array, depth_dimension)

            # Extract just the variables with these spatial coordinates
            dataset_subset = utils.extract_vars(dataset, variable_names)

            # Drop any coordinates for this depth variable.
            # For some reason .isel() call will play havok with them,
            # so best to drop them beforehand.
            dataset_subset = dataset_subset.drop_vars([
                name for name, coordinate in dataset_subset.coords.items()
                if coordinate.dims == (depth_dimension,)
            ])

            # Find the ocean floor using the ocean_floor_indices
            dataset_subset = dataset_subset.isel(
                {depth_dimension: ocean_floor_indices},
                drop=True, missing_dims='ignore')

            # Merge these floored variables back in to the dataset
            dataset = dataset_subset.merge(dataset, compat='override')

    # Finally, drop the depth dimensions.
    # This will clear up any residual variables that use the depth variables,
    # such as depth coordinate variables.
    # errors='ignore' because the depth dimensions may have already been dropped
    dataset = dataset.drop_dims(depth_dimensions, errors='ignore')

    return dataset


def _find_ocean_floor_indices(
    data_array: xr.DataArray,
    depth_dimension: Hashable,
) -> xr.DataArray:
    # This needs some explaining.
    # (any number * 0 + 1) is 1, while (nan * 0 + 1) is nan.
    # As layers under the ocean floor are nans,
    # this will give us a series of 1's for water layers, with nans below,
    # such as [1, 1, 1, nan, nan, nan].
    # `.cumsum()` will then add all the 1's up cumulatively,
    # giving something like `[1, 2, 3, nan, nan, nan]` for a column.
    # `.argmax()` will find the highest non-nan index, and we have our answer!
    #
    # Columns of all nans will have an argmax index of 0.
    # Item 0 in the column will be nan, resulting in nan in the output as desired.
    depth_indices = (data_array * 0 + 1).cumsum(depth_dimension)
    max_depth_indices = depth_indices.argmax(depth_dimension)
    return cast(xr.DataArray, max_depth_indices)


def normalize_depth_variables(
    dataset: xr.Dataset,
    depth_variables: List[Hashable],
    *,
    positive_down: bool = True,
    deep_to_shallow: bool = True,
) -> xr.Dataset:
    """
    Some datasets represent depth as a positive variable, some as negative.
    Some datasets sort depth layers from deepest to most shallow, others
    from shallow to deep. :func:`normalize_depth_variables` will return
    a new dataset with the depth variables normalized.

    The default behaviour is for positive values to indicate deeper depths
    (indicated via the variable attribute ``positive: "down"``),
    and for the layers to be ordered deep to shallow.
    This behaviour can be modified using the parameters
    ``positive_down`` and ``deep_to_shallow`` respectively.

    Parameters
    ----------
    dataset
        The dataset to normalize
    depth_variables
        The names of the depth coordinate variables.
        This should be the names of the variables, not the dimensions,
        for datasets where these differ.
    positive_down
        When true (the default), positive values will indicate depth below the
        surface. When false, negative values indicate depth below the surface.
    deep_to_shallow
        When true (the default), the layers are ordered such that deeper layers
        have lower indices.

    See also
    --------

    :meth:`.Convention.normalize_depth_variables`
    :meth:`.Convention.get_all_depth_names`
    """
    new_dataset = dataset.copy()
    for name in depth_variables:
        variable = dataset[name]
        if len(variable.dims) != 1:
            raise ValueError(
                f"Can't normalize multidimensional depth variable {name!r} "
                f"with dimensions {list(variable.dims)!r}"
            )
        dimension = variable.dims[0]

        new_variable = new_dataset[name]
        new_variable.attrs['positive'] = 'down' if positive_down else 'up'

        positive_attr = variable.attrs.get('positive')
        if positive_attr is None:
            # This is a _depth_ variable. If there are more values >0 than <0,
            # positive is probably down.
            total_values = len(variable.values)
            positive_values = len(variable.values[variable.values > 0])
            positive_attr = 'down' if positive_values > total_values / 2 else 'up'

            warnings.warn(
                f"Depth variable {name!r} had no 'positive' attribute, "
                f"guessing `positive: {positive_attr!r}`",
                stacklevel=2)

        # Reverse the polarity
        if (positive_attr == 'down') != positive_down:
            new_values = -1 * new_variable.values
            if name == dimension:
                new_dataset = new_dataset.assign_coords({name: new_values})
                new_dataset[name].attrs = new_variable.attrs
                new_dataset[name].encoding = new_variable.encoding
                new_variable = new_dataset[name]
            else:
                new_dataset = new_dataset.assign({
                    name: ([dimension], new_values, new_variable.attrs, new_variable.encoding)
                })
                new_variable = new_dataset[name]

        # Check if the existing data goes from deep to shallow, correcting for
        # the positive_down we just adjusted above. This assumes that depth
        # data are monotonic across all values. If this is not the case,
        # good luck.
        d1, d2 = new_variable.values[0:2]
        data_deep_to_shallow = (d1 > d2) == positive_down

        # Flip the order of the coordinate
        if data_deep_to_shallow != deep_to_shallow:
            new_dataset = new_dataset.isel({dimension: np.s_[::-1]})

    return new_dataset



import argparse
import logging
from pathlib import Path
from typing import Callable, Dict

import xarray as xr

import emsarray
from emsarray.cli import BaseCommand, CommandException
from emsarray.operations import geometry
from emsarray.types import Pathish
from emsarray.utils import PerfTimer

logger = logging.getLogger(__name__)

Writer = Callable[[xr.Dataset, Pathish], None]
format_writers: Dict[str, Writer] = {
    'geojson': geometry.write_geojson,
    'shapefile': geometry.write_shapefile,
    'wkt': geometry.write_wkt,
    'wkb': geometry.write_wkb,
}


class Command(BaseCommand):
    help = "Export the geometry of a dataset"
    description = """
    Export the geometry of a dataset to a file
    """

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "input_path", type=Path,
            metavar='input-dataset',
            help="Path to input netCDF4 file")
        parser.add_argument(
            'output_path', type=Path,
            metavar='output',
            help=("Path to exported geometry"))

        parser.add_argument(
            "-f", "--format", type=str,
            default="auto",
            choices=("auto", "geojson", "wkt", "wkb", "shapefile"),
            help=(
                "Format for exported geometry. "
                "The output format will be guessed using the output extension by default"
            ))

    def guess_format(self, output_path: Path) -> str:
        extension = output_path.suffix
        if extension in {'.json', '.geojson'}:
            return 'geojson'
        if extension == '.wkt':
            return 'wkt'
        if extension == '.wkb':
            return 'wkb'
        if extension == '.shp':
            return 'shapefile'
        raise CommandException(
            f"Could not guess output format from extension {extension!r}")

    def handle(self, options: argparse.Namespace) -> None:
        logger.info("Exporting geometry for %r", str(options.input_path))
        dataset = emsarray.open_dataset(options.input_path)

        output_path: Path = options.output_path
        output_format: str = options.format
        if output_format == 'auto':
            output_format = self.guess_format(output_path)
            logger.debug("Guessed output format as %r", output_format)

        with PerfTimer() as t:
            count = dataset.ems.polygons.size
        logger.debug("Generated %d polygons in %f seconds", count, t.elapsed)

        try:
            writer = format_writers[output_format]
        except KeyError:
            raise CommandException(f"Unknown output format {output_format!r}")

        logger.debug("Exporting geometry as %r", output_format)
        writer(dataset, output_path)



import importlib
import sys

import pytest
import xarray as xr


@pytest.fixture(autouse=True)
def remove_cached_module():
    sys.modules.pop('emsarray.formats', None)


def test_warning_on_import():
    with pytest.warns(DeprecationWarning):
        importlib.import_module('emsarray.formats')


def test_warning_on_get_file_format(datasets):
    with pytest.warns(DeprecationWarning):
        from emsarray.formats import get_file_format

    dataset = xr.open_dataset(datasets / 'cfgrid1d.nc')
    with pytest.warns(DeprecationWarning):
        get_file_format(dataset)


def test_warning_on_format_subclass(datasets):
    with pytest.warns(DeprecationWarning):
        from emsarray.formats import Format

    with pytest.warns(DeprecationWarning):
        class NewFormat(Format):
            pass



"""
Dataclass for containing state required for emsarray
"""
from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Final, Optional, cast

import xarray as xr

if TYPE_CHECKING:
    from emsarray.conventions._base import Convention


@dataclasses.dataclass
class State:
    """
    Keeps state for emsarray.
    Currently only used to allow binding Convention instances to datasets
    to avoid convention autodetection.
    """
    dataset: xr.Dataset
    convention: Optional[Convention] = None

    accessor_name: Final[str] = "_emsarray_state"

    @classmethod
    def get(cls, dataset: xr.Dataset) -> "State":
        """
        Get the state for a dataset,
        making an empty state if none exists.
        """
        return cast(State, getattr(dataset, State.accessor_name))

    def bind_convention(self, convention: Convention) -> None:
        """
        Bind a Convention instance to this Dataset.
        If the Dataset is already bound, an error is raised.
        """
        self.convention = convention

    def is_bound(self) -> bool:
        """
        Check if the Dataset has a bound convention.
        """
        return self.convention is not None



"""
Exception and warning classes for specific exceptions
relating to emsarray and datasets.
"""


class EmsarrayError(Exception):
    """
    Base class for all emsarray-specific exception classes.
    """


class ConventionViolationError(EmsarrayError):
    """
    A dataset violates its conventions in a way that is not recoverable.
    """


class ConventionViolationWarning(UserWarning):
    """
    A dataset violates its conventions in a way that we can handle.
    For example, an attribute has an invalid type,
    but is still interpretable.
    """



import logging
from typing import Any

import xarray as xr

from emsarray.types import Pathish

logger = logging.getLogger(__name__)


def open_dataset(path: Pathish, **kwargs: Any) -> xr.Dataset:
    """
    Open a dataset and determine the correct Convention implementation for it.
    If a valid Convention implementation can not be found, an error is raised.

    Parameters
    ----------
    path
        The path to the dataset to open
    kwargs
        These are passed straight through to :func:`xarray.open_dataset`.

    Returns
    -------
    xarray.Dataset
        The opened dataset

    Example
    -------

    .. code-block:: python

        import emsarray
        dataset = emsarray.open_dataset("./tests/datasets/ugrid_mesh2d.nc")

    See also
    --------
    :func:`xarray.open_dataset`
    """
    dataset = xr.open_dataset(path, **kwargs)

    # Determine the correct convention. All the magic happens in the accessor.
    convention = dataset.ems
    convention_class = type(convention)
    logger.debug(
        "Using convention %s.%s for dataset %r",
        convention_class.__module__, convention_class.__name__, str(path))

    return dataset



"""
Unstructured grid convention.

See also
--------
`UGRID conventions <https://ugrid-conventions.github.io/ugrid-conventions/>`_
"""
from __future__ import annotations

import enum
import logging
import pathlib
import warnings
from collections import defaultdict
from contextlib import suppress
from dataclasses import dataclass
from functools import cached_property
from typing import (
    Any, Dict, FrozenSet, Hashable, Iterable, List, Mapping, Optional, Set,
    Tuple, cast
)

import numpy as np
import shapely
import xarray as xr
from shapely.geometry.base import BaseGeometry

from emsarray import utils
from emsarray.exceptions import (
    ConventionViolationError, ConventionViolationWarning
)
from emsarray.types import Pathish

from ._base import Convention, Specificity

logger = logging.getLogger(__name__)


def _split_coord(attr: str) -> Tuple[str, str]:
    x, y = attr.split(None, 1)
    return (x, y)


def buffer_faces(face_indices: np.ndarray, topology: Mesh2DTopology) -> np.ndarray:
    """
    When clipping a dataset to a region, including a buffer of extra faces
    around the included faces is desired. Given an array of face indices,
    :func:`buffer_faces` will find all the faces required to form this buffer
    and return a new array of face indices that include both the input faces
    and the buffer faces

    Specifically, this finds all faces that share a node with one of the
    included faces.
    """
    original_face_indices = set(face_indices.tolist())
    face_node = topology.face_node_array

    # Find all nodes for the faces
    included_nodes = set(np.unique(face_node[face_indices].compressed()))

    # Find all faces that are composed of any of the nodes just found
    included_faces = (
        face_index
        for face_index, node_indices in enumerate(face_node)
        # Either one of the original faces ...
        if face_index in original_face_indices
        # ... or shares a node with one of the original faces
        or bool(included_nodes.intersection(node_indices.compressed()))
    )
    return cast(np.ndarray, np.fromiter(included_faces, dtype=topology.sensible_dtype))


def mask_from_face_indices(face_indices: np.ndarray, topology: Mesh2DTopology) -> xr.Dataset:
    """
    Make a mask dataset from a list of face indices.
    This mask can later be applied using :meth:`~.Convention.apply_clip_mask`.

    A mask for a UGRID dataset indicates which nodes, edges, and faces
    (collectively 'elements') to include, and their new indices. When
    elements are dropped from the dataset, to keep the remaining data
    contiguous, all elements gain new indices.

    The mask consists of three data arrays corresponding to nodes, edges,
    and faces. Each data array indicates the new index for that element, or
    whether that element should be dropped.
    """
    # Generate a new index for each element. Indices are assigned in the same
    # order as they currently exist. This means that data rows for excluded
    # elements can be dropped, and the remaining data rows will automatically
    # have the correct indices.
    fill_value = topology.sensible_fill_value
    data_vars = {}

    def new_element_indices(size: int, indices: np.ndarray) -> np.ma.MaskedArray:
        new_indices = np.full(
            (size,), fill_value=fill_value, dtype=topology.sensible_dtype)
        new_indices = np.ma.masked_array(new_indices, mask=True)
        new_indices[indices] = np.arange(len(indices))
        return new_indices

    # Record which old face index maps to which new face index
    data_vars['new_face_index'] = _masked_integer_data_array(
        data=new_element_indices(topology.face_count, face_indices),
        fill_value=fill_value,
        dims=['old_face_index'],
    )

    # Find all edges associated with included faces. Don't bother if the
    # dataset doesn't define an edge dimension
    if topology.has_edge_dimension:
        face_edge = topology.face_edge_array
        edge_indices = np.sort(np.unique(face_edge[face_indices].compressed()))

        # Record which old edge index maps to which new edge index
        data_vars['new_edge_index'] = _masked_integer_data_array(
            data=new_element_indices(topology.edge_count, edge_indices),
            fill_value=fill_value,
            dims=['old_edge_index'],
        )

    # Find all nodes associated with included faces
    face_node = topology.face_node_array
    node_indices = np.sort(np.unique(face_node[face_indices].compressed()))

    # Record which old node index maps to which new node index
    data_vars['new_node_index'] = _masked_integer_data_array(
        data=new_element_indices(topology.node_count, node_indices),
        fill_value=fill_value,
        dims=['old_node_index'],
    )

    # Make the mask dataset
    return xr.Dataset(data_vars=data_vars, attrs={
        'title': 'UGRID dataset mask',
    })


def _get_start_index(connectivity: xr.DataArray) -> int:
    """
    Get the ``start_index`` attribute from a connectivity data array,
    while checking for common error cases.
    """
    # Default to 0 if there is no start_index attribute
    if 'start_index' not in connectivity.attrs:
        return 0

    start_index = connectivity.attrs.get('start_index')

    # These are the valid values
    if start_index in {0, 1}:
        return cast(int, start_index)

    # Some datasets use the strings '0' or '1'.
    # This does not adhere to the spec but is easy to interpret
    if start_index in {'0', '1'}:
        warnings.warn(
            f"Connectivity data array {connectivity.name!r} "
            f"had string-typed start index {start_index!r}. "
            f"This attribute should be the integer {int(start_index)}, "
            f"not a string.",
            ConventionViolationWarning,
        )
        return int(start_index)

    # Anything else is incorrect
    raise ConventionViolationError(
        f"Connectivity data array {connectivity.name!r} "
        f"had invalid start index {start_index!r}"
    )


def update_connectivity(
    connectivity: xr.DataArray,
    old_array: np.ndarray,
    row_indices: np.ndarray,
    column_values: np.ndarray,
    primary_dimension: Hashable,
    fill_value: int,
) -> xr.DataArray:
    """
    Create a new connectivity variable by reindexing existing entries.
    This is used during masking to trim off unused nodes, edges, and faces.
    It preserves original dtype, any given ``start_index`` value,
    and non-standard dimension orderings.

    Parameters
    ----------

    connectivity : xarray.DataArray
        The connectivity variable to update.
        This will be one of
        :attr:`~Mesh2DTopology.edge_node_connectivity`,
        :attr:`~Mesh2DTopology.edge_face_connectivity`,
        :attr:`~Mesh2DTopology.face_node_connectivity`,
        :attr:`~Mesh2DTopology.face_edge_connectivity`, or
        :attr:`~Mesh2DTopology.face_face_connectivity`.
    old_array : np.ndarray
        The old connectivity array,
        the companion to the connectivity data array.
        Each row corresponds to either an edge or a face.
    row_indices : np.ndarray
        A one dimensional numpy masked array
        indicating which rows of ``old_array`` are to be included.
        Masked items are excluded, all other items are included.
    column_values : np.ndarray
        A one dimensional numpy masked array
        mapping values from ``old_array`` to their new values.
        Values to be excluded in the new array should be ``np.ma.masked``
    primary_dimension : Hashable
        The name of the primary dimension for this connectivity variable.
        This will be either :attr:`Mesh2DTopology.edge_dimension`
        or :attr:`Mesh2DTopology.face_dimension`.
        Each row represents data about either a edge or a face.
    fill_value : int
        The fill value to use for missing data.

    Returns
    -------

    xarray.DataArray
        A new DataArray containing the clipped and reindexed connectivity data.
    """
    logger.debug("Reindexing %r", connectivity.name)

    # Offset the column_values for one-based indexing.
    start_index = _get_start_index(connectivity)
    if start_index != 0:
        column_values = column_values + start_index

    # We need to preseve the integer dtype,
    # while also accounting for masked values.
    # xarray does not make this easy.
    # By constructing the array using new_fill_value where needed,
    # setting the dtype explicitly, and adding the _FillValue attribute,
    # xarray will cooperate.
    dtype = connectivity.encoding.get('dtype', connectivity.dtype)
    include_row = ~np.ma.getmask(row_indices)
    raw_values = np.array([
        [
            column_values[item] if item is not np.ma.masked else fill_value
            for item in row
        ]
        for row in old_array[include_row]
    ], dtype=dtype)
    values = np.ma.masked_equal(raw_values, fill_value)

    if connectivity.dims[1] == primary_dimension:
        values = np.transpose(values)
    elif primary_dimension not in connectivity.dims:
        raise ValueError("Connectivity variable does not contain primary dimension")

    return _masked_integer_data_array(
        data=values,
        fill_value=fill_value,
        dims=connectivity.dims,
        name=connectivity.name,
        attrs=connectivity.attrs,
    )


def _masked_integer_data_array(
    data: np.ma.MaskedArray,
    fill_value: int,
    **kwargs: Any,
) -> xr.DataArray:
    """
    Create an :class:`xarray.DataArray` that represents
    an integer variable with a _FillValue.
    When reading such a variable from disk,
    xarray will cast it to a double
    and replace any occurrence of _FillValue with np.nan.
    This method replicates these actions.
    When saved to disk, a variable built using this should have
    an integer type and a _FillValue.
    """
    float_data = np.ma.filled(data.astype(np.double), np.nan)
    data_array = xr.DataArray(data=float_data, **kwargs)
    data_array.encoding.update({'dtype': data.dtype, '_FillValue': fill_value})
    return data_array


class NoEdgeDimensionException(ValueError):
    """
    Raised when the dataset does not define a name for the edge dimension, and
    no dimension name can be derived from the available connectivity variables.
    """


class NoConnectivityVariableException(Exception):
    """
    Raised when the dataset does not have a particular connectivity variable
    defined.
    """


@dataclass
class Mesh2DTopology:
    """
    A helper for extracting the topology data for a 2D unstructured mesh
    dataset. The dataset will contain a dummy variable with the attribute
    ``cf_role = "mesh_topology"``. The attributes of that variable explain the
    topology of the dataset. For example, the names of the variables that store
    the coordinates for nodes can be found in the ``node_coordinates`` attribute
    as the string ``"x_variable y_variable"``.

    There are five possible connectivity variables on a UGRID dataset:
    face-node, face-edge, face-face, edge-node, and edge-face.
    Face-node is the only required variable.
    Each has a property to check whether it is defined (``has_valid_foo``),
    a property to get the variable if it is defined (``foo_connectivity``),
    and a property to get or derive a normalised connectivity array (``foo_array``):

    .. list-table::
        :header-rows: 1

        * - Attribute
          - has_valid_foo
          - foo_connectivity
          - foo_array
        * - face_node
          - :attr:`has_valid_face_node_connectivity`
          - :attr:`face_node_connectivity`
          - :attr:`face_node_array`
        * - face_edge
          - :attr:`has_valid_face_edge_connectivity`
          - :attr:`face_edge_connectivity`
          - :attr:`face_edge_array`
        * - face_face
          - :attr:`has_valid_face_face_connectivity`
          - :attr:`face_face_connectivity`
          - :attr:`face_face_array`
        * - edge_node
          - :attr:`has_valid_edge_node_connectivity`
          - :attr:`edge_node_connectivity`
          - :attr:`edge_node_array`
        * - edge_face
          - :attr:`has_valid_edge_face_connectivity`
          - :attr:`edge_face_connectivity`
          - :attr:`edge_face_array`
    """
    #: The UGRID dataset
    dataset: xr.Dataset

    #: The name of the mesh topology variable. Optional. If not provided, the
    #: mesh topology dummy variable will be found by checking the ``cf_role``
    #: attribute.
    topology_key: Optional[Hashable] = None

    #: The default dtype to use for index data arrays. Hard coded to ``int32``,
    #: which should be sufficient for all datasets. ``int16`` is too small for
    #: many datasets, even after slicing.
    sensible_dtype = np.int32

    def __repr__(self) -> str:
        attrs = {
            'nodes': self.node_count,
            'edges': self.edge_count if self.has_edge_dimension else 'na',
            'faces': self.face_count,
        }
        attr_str = ' '.join(f'{key}: {value!r}' for key, value in attrs.items())
        return f'<{type(self).__name__} {attr_str}>'

    @cached_property
    def mesh_variable(self) -> xr.DataArray:
        """
        Get the dummy variable that stores the mesh topology information.
        This is variable is either named using the ``topology_key`` argument to
        the contructor, or found by looking for a variable with the attribute
        ``cf_role`` of ``"mesh_topology"``.
        """
        if self.topology_key is not None:
            return self.dataset.data_vars[self.topology_key]
        try:
            return next(
                data_array for data_array in self.dataset.data_vars.values()
                if data_array.attrs.get('cf_role') == 'mesh_topology'
            )
        except StopIteration:
            raise ValueError("No mesh variable found")

    @property
    def mesh_attributes(self) -> Dict[Hashable, str]:
        """
        Get the mesh topology attributes from the dummy variable with the
        attribute ``cf_role`` of ``"mesh_topology"``.
        """
        return {
            key: value for key, value in self.mesh_variable.attrs.items()
            if isinstance(value, str)
        }

    @cached_property
    def sensible_fill_value(self) -> int:
        """
        A sensible value to use for ``_FillValue``.
        This is all-nines number larger than each of
        :attr:`node_count`, :attr:`edge_count`, and :attr:`face_count`.

        An alternate, simpler implementation would use ``np.iinfo(dtype).max``
        (the maximum integer value for a given dtype),
        but using all-nines is traditional.
        """
        # There will always be more edges than faces, as one face is composed
        # of at least three edges. There are probably more edges than nodes.
        # Not all datasets define how many edges they contain, and computing
        # that can be expensive. The upper bound for the number of edges in the
        # dataset is face_count * max_node_count, which is used as an estimate.
        # This is the highest possible index used in the dataset
        max_count = max([self.node_count, self.face_count * self.max_node_count])

        # This way is more mathematically fun, but less performant
        # fill_value =  10 ** (math.floor(math.log10(max_count)) + 2) - 1
        return int('9' * (len(str(max_count)) + 1))

    @cached_property
    def _node_coordinates(self) -> Tuple[Hashable, Hashable]:
        return _split_coord(self.mesh_attributes['node_coordinates'])

    @cached_property
    def _edge_coordinates(self) -> Tuple[Hashable, Hashable]:
        return _split_coord(self.mesh_attributes['edge_coordinates'])

    @cached_property
    def _face_coordinates(self) -> Tuple[Hashable, Hashable]:
        return _split_coord(self.mesh_attributes['face_coordinates'])

    @property
    def node_x(self) -> xr.DataArray:
        """Data array of node X / longitude coordinates."""
        return self.dataset.data_vars[self._node_coordinates[0]]

    @property
    def node_y(self) -> xr.DataArray:
        """Data array of node Y / latitude coordinates."""
        return self.dataset.data_vars[self._node_coordinates[1]]

    @property
    def edge_x(self) -> Optional[xr.DataArray]:
        """Data array of characteristic edge X / longitude coordinates. Optional."""
        try:
            return self.dataset.data_vars[self._edge_coordinates[0]]
        except KeyError:
            return None

    @property
    def edge_y(self) -> Optional[xr.DataArray]:
        """Data array of characteristic edge y / latitude coordinates. Optional."""
        try:
            return self.dataset.data_vars[self._edge_coordinates[1]]
        except KeyError:
            return None

    @property
    def face_x(self) -> Optional[xr.DataArray]:
        """Data array of characteristic face x / longitude coordinates. Optional."""
        try:
            return self.dataset.data_vars[self._face_coordinates[0]]
        except KeyError:
            return None

    @property
    def face_y(self) -> Optional[xr.DataArray]:
        """Data array of characteristic face y / latitude coordinates. Optional."""
        try:
            return self.dataset.data_vars[self._face_coordinates[1]]
        except KeyError:
            return None

    def _to_index_array(
        self,
        data_array: xr.DataArray,
        primary_dimension: Hashable,
    ) -> np.ndarray:
        """
        Convert a data array of node, edge, or face indices
        into a masked numpy integer array.
        This takes care of casting arrays to integers
        and fixing one-based indexing where appropriate.
        """
        # Ensure the dimensions are in the correct order first
        if primary_dimension not in data_array.dims:
            raise ConventionViolationError(
                f"Data array {data_array.name!r} did not contain "
                f"primary dimension {primary_dimension!r}")
        if data_array.dims[0] != primary_dimension:
            data_array = data_array.transpose()

        values = data_array.values

        if not issubclass(values.dtype.type, np.integer):
            # If a data array has a fill value, xarray will convert that data array
            # to a floating point data type, and replace masked values with np.nan.
            # Here we convert a floating point array to a masked integer array.
            values = np.ma.masked_invalid(values).astype(np.int_)
        elif '_FillValue' in data_array.attrs:
            # The data array has a fill value, but xarray has not applied it.
            # This implied the dataset was opened with mask_and_scale=False,
            # or was constructed in memory.
            # Convert it to a mask array with _FillValue masked out.
            values = np.ma.masked_equal(values, data_array.attrs['_FillValue'])
        else:
            # If the value is still an integer then it had no fill value.
            # Convert it to a mask array with no masked values.
            values = np.ma.masked_array(values, mask=np.ma.nomask)

        # UGRID conventions allow for zero based or one based indexing.
        # To be consistent we convert all indices to zero based.
        start_index = _get_start_index(data_array)
        if start_index != 0:
            values = values - start_index

        return values

    @cached_property
    def has_valid_edge_node_connectivity(self) -> bool:
        """
        Does the dataset contain a valid edge_node_connectivity variable?
        Some datasets have an attribute naming a edge_node_connectivity
        variable, but the variable doesn't exist. Other times, the shape of the
        variable is incorrect.
        """
        if not self.has_edge_dimension:
            return False

        try:
            data_array = self.dataset.data_vars[self.mesh_attributes["edge_node_connectivity"]]
        except KeyError:
            return False

        actual = set(data_array.dims)
        expected = {self.edge_dimension, self.two_dimension}
        if actual != expected:
            warnings.warn(
                f"Got a edge_node_connectivity variable {data_array.name!r} with "
                f"unexpected dimensions {actual}, expecting {expected}",
                ConventionViolationWarning,
            )
            return False

        return True

    @cached_property
    def edge_node_connectivity(self) -> xr.DataArray:
        """
        This data array defines unique indexes for each edge. This allows data
        to be stored 'on' an edge.

        If this variable is *not* defined,
        data are not stored on the edges for this dataset.
        """
        if not self.has_edge_dimension:
            raise NoEdgeDimensionException

        if not self.has_valid_edge_node_connectivity:
            raise NoConnectivityVariableException(
                "No valid edge_node_connectivity defined")

        name = self.mesh_attributes['edge_node_connectivity']
        return self.dataset.data_vars[name]

    @cached_property
    def edge_node_array(self) -> np.ndarray:
        """
        An integer numpy array with shape
        (:attr:`edge_dimension`, 2),
        indicating which nodes each edge consists of.
        If data are stored on edges in this dataset,
        the ordering of these edges matches the ordering of the data.

        If this variable is *not* defined,
        we can derive it from the existing data,
        arbitrarily assigning an index to each edge.
        This will allow us to derive other connectivity variables if required.
        """
        if not self.has_edge_dimension:
            raise NoEdgeDimensionException

        if self.has_valid_edge_node_connectivity:
            return self._to_index_array(
                self.edge_node_connectivity, self.edge_dimension)

        logger.info("Building edge_node_array")
        with utils.PerfTimer() as timer:
            # Each edge is composed of two nodes. Each edge may be named twice,
            # once for each face. To de-duplicate this, edges are built up using
            # this dict-of-sets, where the dict index is the node with the
            # lower index, and the set is the node indices of the other end.
            low_highs: Dict[int, Set[int]] = defaultdict(set)

            for face_index, node_pairs in self._face_and_node_pair_iter():
                for pair in node_pairs:
                    low, high = sorted(pair)
                    low_highs[low].add(high)
            edge_node = np.array([
                [low, high]
                for low, highs in low_highs.items()
                for high in highs
            ], dtype=self.sensible_dtype)
        logger.debug("Built edge_node_array in %f seconds", timer.elapsed)

        return edge_node

    @cached_property
    def has_valid_edge_face_connectivity(self) -> bool:
        """
        Does the dataset contain a valid edge_face_connectivity variable?
        Some datasets have an attribute naming a edge_face_connectivity
        variable, but the variable doesn't exist. Other times, the shape of the
        variable is incorrect.
        """
        if not self.has_edge_dimension:
            return False

        try:
            data_array = self.dataset.data_vars[self.mesh_attributes["edge_face_connectivity"]]
        except KeyError:
            return False

        actual = set(data_array.dims)
        expected = {self.edge_dimension, self.two_dimension}
        if actual != expected:
            warnings.warn(
                f"Got a face_face_connectivity variable {data_array.name!r} with "
                f"unexpected dimensions {actual}, expecting {expected}",
                ConventionViolationWarning,
            )
            return False

        return True

    @cached_property
    def edge_face_connectivity(self) -> xr.DataArray:
        """
        This data array shows which faces an edge borders on.
        """
        if not self.has_valid_edge_face_connectivity:
            raise NoConnectivityVariableException(
                "No valid edge_face_connectivity defined")
        name = self.mesh_attributes['edge_face_connectivity']
        return self.dataset.data_vars[name]

    @cached_property
    def edge_face_array(self) -> np.ndarray:
        """
        An integer numpy array with shape
        (:attr:`edge_dimension`, 2),
        indicating which faces border each edge.
        """
        if self.has_valid_edge_face_connectivity:
            return self._to_index_array(
                self.edge_face_connectivity, self.edge_dimension)

        # Access these outside of the timer below
        fill_value = self.sensible_fill_value
        face_edge = self.face_edge_array

        # Build an edge_face_connectivity matrix
        logger.info("Building edge_face_array")
        with utils.PerfTimer() as timer:
            # The edge_face connectivity matrix
            shape = (self.edge_count, 2)
            filled = np.full(shape, fill_value, dtype=self.sensible_dtype)
            edge_face: np.ndarray = np.ma.masked_array(filled, mask=True)

            # The number of faces already seen for this edge
            edge_face_count = np.zeros(self.edge_count, dtype=self.sensible_dtype)

            for face_index, edge_indices in enumerate(face_edge):
                for edge_index in edge_indices.compressed():
                    edge_face[edge_index, edge_face_count[edge_index]] = face_index
                    edge_face_count[edge_index] += 1
        logger.debug("Built edge_face_array in %f seconds", timer.elapsed)

        return edge_face

    @cached_property
    def has_valid_face_node_connectivity(self) -> bool:
        """
        Does the dataset contain a valid face_node_connectivity variable?
        If this is invalid, the entire dataset is invalid.
        """
        try:
            data_array = self.dataset.data_vars[self.mesh_attributes["face_node_connectivity"]]
        except KeyError:
            return False

        actual = set(data_array.dims)
        expected = {self.face_dimension, self.max_node_dimension}
        if actual != expected:
            warnings.warn(
                f"Got a face_node_connectivity variable {data_array.name!r} with "
                f"unexpected dimensions {actual}, expecting {expected}",
                ConventionViolationWarning,
            )
            return False

        return True

    @cached_property
    def face_node_connectivity(self) -> xr.DataArray:
        """
        A variable that lists the nodes that make up the boundary of each face.
        This is the only required data variable in a UGRID dataset,
        all the others can be derived from this if required.
        """
        name = self.mesh_attributes['face_node_connectivity']
        return self.dataset.data_vars[name]

    @cached_property
    def face_node_array(self) -> np.ndarray:
        """
        An integer numpy array with shape
        (:attr:`face_dimension`, :attr:`max_node_dimension`),
        representing the node indices that make up each face.
        """
        return self._to_index_array(
            self.face_node_connectivity, self.face_dimension)

    @cached_property
    def has_valid_face_edge_connectivity(self) -> bool:
        """
        Does the dataset contain a valid face_edge_connectivity variable?
        Some datasets have an attribute naming a face_edge_connectivity
        variable, but the variable doesn't exist. Other times, the shape of the
        variable is incorrect.
        """
        try:
            data_array = self.dataset.data_vars[self.mesh_attributes["face_edge_connectivity"]]
        except KeyError:
            return False

        actual = set(data_array.dims)
        expected = {self.face_dimension, self.max_node_dimension}
        if actual != expected:
            warnings.warn(
                f"Got a face_edge_connectivity variable {data_array.name!r} with "
                f"unexpected dimensions {actual}, expecting {expected}",
                ConventionViolationWarning,
            )
            return False

        return True

    @cached_property
    def face_edge_connectivity(self) -> xr.DataArray:
        """
        The face_edge_connectivity variable from the dataset, if present.
        """
        if not self.has_valid_face_edge_connectivity:
            raise NoConnectivityVariableException(
                "No valid face_edge_connectivity defined")
        name = self.mesh_attributes['face_edge_connectivity']
        return self.dataset.data_vars[name]

    @cached_property
    def face_edge_array(self) -> np.ndarray:
        """
        An integer numpy array with shape
        (:attr:`face_dimension`, :attr:`max_node_dimension`),
        representing the edge indices that border each face.
        """
        if self.has_valid_face_edge_connectivity:
            return self._to_index_array(
                self.face_edge_connectivity, self.face_dimension)

        # Access these outside of the timer below
        fill_value = self.sensible_fill_value
        edge_node = self.edge_node_array

        # Build a face_edge_connectivity matrix
        logger.info("Building face_edge_array")
        with utils.PerfTimer() as timer:
            shape = (self.face_count, self.max_node_count)
            filled = np.full(shape, fill_value, dtype=self.sensible_dtype)
            face_edge: np.ndarray = np.ma.masked_array(filled, mask=True)

            node_pair_to_edge_index = {
                frozenset(edge): edge_index
                for edge_index, edge in enumerate(edge_node)
            }

            for face_index, node_pairs in self._face_and_node_pair_iter():
                for column, node_pair in enumerate(node_pairs):
                    edge_index = node_pair_to_edge_index[frozenset(node_pair)]
                    face_edge[face_index, column] = edge_index
        logger.debug("Built face_edge_array in %f seconds", timer.elapsed)

        return face_edge

    @cached_property
    def has_valid_face_face_connectivity(self) -> bool:
        """
        Does the dataset contain a valid face_face_connectivity variable?
        Some datasets have an attribute naming a face_face_connectivity
        variable, but the variable doesn't exist. Other times, the shape of the
        variable is incorrect.
        """
        try:
            data_array = self.dataset.data_vars[self.mesh_attributes["face_face_connectivity"]]
        except KeyError:
            return False

        actual = set(data_array.dims)
        expected = {self.face_dimension, self.max_node_dimension}
        if actual != expected:
            warnings.warn(
                f"Got a face_face_connectivity variable {data_array.name!r} with "
                f"unexpected dimensions {actual}, expecting {expected}",
                ConventionViolationWarning,
            )
            return False

        return True

    @cached_property
    def face_face_connectivity(self) -> xr.DataArray:
        """
        The face_face_connectivity variable from the dataset, if present.
        """
        if not self.has_valid_face_face_connectivity:
            raise NoConnectivityVariableException(
                "No valid face_face_connectivity defined")
        name = self.mesh_attributes['face_face_connectivity']
        return self.dataset.data_vars[name]

    @cached_property
    def face_face_array(self) -> np.ndarray:
        """
        An integer numpy array with shape
        (:attr:`face_dimension`, :attr:`max_node_dimension`),
        representing the indices of all faces that border a given face.
        """
        if self.has_valid_face_face_connectivity:
            return self._to_index_array(
                self.face_face_connectivity, self.face_dimension)

        # Access these outside of the timer below
        fill_value = self.sensible_fill_value
        edge_face = self.edge_face_array

        # Build a face_face_connectivity matrix
        logger.info("Building face_face_array")
        with utils.PerfTimer() as timer:
            face_count = np.zeros(self.face_count, dtype=self.sensible_dtype)
            shape = (self.face_count, self.max_node_count)
            filled = np.full(shape, fill_value, dtype=self.sensible_dtype)
            face_face: np.ndarray = np.ma.masked_array(filled, mask=True)

            for edge_index, face_indices in enumerate(edge_face):
                if np.any(np.ma.getmask(face_indices)):
                    continue
                left, right = face_indices
                face_face[left, face_count[left]] = right
                face_face[right, face_count[right]] = left
                face_count[left] += 1
                face_count[right] += 1
        logger.debug("Built face_face_array in %f seconds", timer.elapsed)

        return face_face

    def _face_and_node_pair_iter(self) -> Iterable[Tuple[int, List[Tuple[int, int]]]]:
        """
        An iterator returning a tuple of ``(face_index, edges)``,
        where ``edges`` is a list of ``(node_index, node_index)`` tuples
        defining the edges of the face.
        """
        face_node = self.face_node_array
        for face_index, node_indices in enumerate(face_node):
            node_indices = node_indices.compressed()
            node_indices = np.append(node_indices, node_indices[0])
            yield face_index, list(utils.pairwise(node_indices))

    @cached_property
    def dimension_for_grid_kind(self) -> Dict[UGridKind, Hashable]:
        """
        Get the dimension names for each of the grid types in this dataset.
        """
        dimensions = {
            UGridKind.face: self.face_dimension,
            UGridKind.node: self.node_dimension,
        }
        if self.has_edge_dimension:
            dimensions[UGridKind.edge] = self.edge_dimension
        return dimensions

    @cached_property
    def two_dimension(self) -> Hashable:
        """
        Get the name of the dimension with size two, for things like edge
        connectivity. The standard name for this dimension is 'Two'.
        """
        two = 'Two'
        # Check for the standard name
        if two in self.dataset.dims and self.dataset.dims[two] == 2:
            return two
        # Check for any other dimension of size 2
        for name, size in self.dataset.dims.items():
            if size == 2:
                return name
        # Make up a new dimension with the standard name
        return two

    @property
    def node_dimension(self) -> Hashable:
        """The name of the dimension for the number of nodes."""
        # This is implicitly given by some of the required variables, such as
        # the variable for node x-coordinates.
        return self.node_x.dims[0]

    @property
    def has_edge_dimension(self) -> bool:
        """
        Check whether this dataset has an edge dimension.
        If not, data are not stored on edges in this dataset.
        """
        if 'edge_dimension' in self.mesh_attributes:
            return True

        # We do not use `self.has_valid_edge_node_connectivity` here to prevent
        # circular logic. This leaves open the possibility of a dataset having
        # an invalid edge_node_connectivity variable without a proper edge
        # dimension defined, but this attribute being True regardless.
        topo_keys = ['edge_node_connectivity', 'edge_face_connectivity']
        return any(
            key in self.mesh_attributes
            and self.mesh_attributes[key] in self.dataset.variables
            for key in topo_keys
        )

    @cached_property
    def edge_dimension(self) -> Hashable:
        """
        The name of the dimension for the number of edges.
        """
        if not self.has_edge_dimension:
            raise NoEdgeDimensionException

        with suppress(KeyError):
            return self.mesh_attributes['edge_dimension']

        # If there is no edge_dimension attribute,
        # the edge dimension is implicitly defined as the first dimension
        # on either of these variables.
        # Both of these variables are optional,
        # leaving the possibility that there is no edge dimension defined.
        topo_keys = ['edge_node_connectivity', 'edge_face_connectivity']
        names = (self.mesh_attributes[key] for key in topo_keys if key in self.mesh_attributes)
        variables = (self.dataset.variables[name] for name in names if name in self.dataset.variables)

        try:
            return next(variable.dims[0] for variable in variables)
        except StopIteration:
            # This should have already happened above, but just in case.
            raise NoEdgeDimensionException

    @cached_property
    def face_dimension(self) -> Hashable:
        """The name of the dimension for the number of faces."""
        try:
            # By definition this is either the dimension named in this attribute ...
            return self.mesh_attributes['face_dimension']
        except KeyError:
            # ... Or the first dimension in this required variable
            return self.face_node_connectivity.dims[0]

    @property
    def max_node_dimension(self) -> Hashable:
        """The name of the dimension for the maximum nodes / edges per face."""
        # This dimension is not named as part of the UGRID spec, but it is
        # always the other dimension in the `face_node_connectivity` variable.
        dims = set(self.face_node_connectivity.dims)
        assert len(dims) == 2
        dims.remove(self.face_dimension)
        return dims.pop()

    @property
    def node_count(self) -> int:
        """The number of nodes in the dataset."""
        return self.dataset.dims[self.node_dimension]

    @property
    def edge_count(self) -> int:
        """The number of edges in the dataset."""
        if not self.has_edge_dimension:
            raise NoEdgeDimensionException()

        # This dimension may not be defined, so ignore KeyErrors. We can
        # compute it below.
        with suppress(KeyError):
            return self.dataset.dims[self.edge_dimension]

        # By computing the edge_node array we can determine how many edges exist
        return self.edge_node_array.shape[0]

    @property
    def face_count(self) -> int:
        """
        The number of faces in the dataset.
        """
        return self.dataset.dims[self.face_dimension]

    @property
    def max_node_count(self) -> int:
        """The maximum number of nodes / edges per face."""
        return self.dataset.dims[self.max_node_dimension]


class UGridKind(str, enum.Enum):
    """UGRID datasets can store data on faces, edges, or nodes."""
    face = 'face'
    edge = 'edge'
    node = 'node'


#: UGRID indices are always single integers, for all index kinds.
UGridIndex = Tuple[UGridKind, int]


class UGrid(Convention[UGridKind, UGridIndex]):
    """A :class:`.Convention` subclass to handle unstructured grid datasets.
    """

    default_grid_kind = UGridKind.face

    @classmethod
    def check_dataset(cls, dataset: xr.Dataset) -> Optional[int]:
        """
        A UGrid dataset needs a global attribute of Conventions = 'UGRID/...',
        and a variable with attribute cf_role = 'mesh_topology'
        and topology_dimension = 2
        """
        conventions = str(dataset.attrs.get('Conventions', ''))
        if 'UGRID' not in conventions:
            return None

        topology = Mesh2DTopology(dataset)
        try:
            mesh = topology.mesh_variable
        except ValueError:
            return None

        if mesh.attrs.get('topology_dimension') != 2:
            return None

        return Specificity.HIGH

    @cached_property
    def topology(self) -> Mesh2DTopology:
        """
        The :class:`Mesh2DTopology` of this dataset - which faces connect with which edges.
        """
        return Mesh2DTopology(self.dataset)

    def ravel_index(self, index: UGridIndex) -> int:
        return index[1]

    def unravel_index(
        self,
        index: int,
        grid_kind: Optional[UGridKind] = None,
    ) -> UGridIndex:
        if grid_kind is None:
            grid_kind = UGridKind.face
        return (grid_kind, index)

    @cached_property
    def grid_kinds(self) -> FrozenSet[UGridKind]:
        items = [UGridKind.face, UGridKind.node]
        # The edge dimension is optional, not all UGRID datasets define it
        if self.topology.has_edge_dimension:
            items.append(UGridKind.edge)
        return frozenset(items)

    def get_grid_kind_and_size(
        self, data_array: xr.DataArray,
    ) -> Tuple[UGridKind, int]:
        if self.topology.face_dimension in data_array.dims:
            return (UGridKind.face, self.topology.face_count)
        if self.topology.has_edge_dimension and self.topology.edge_dimension in data_array.dims:
            return (UGridKind.edge, self.topology.edge_count)
        if self.topology.node_dimension in data_array.dims:
            return (UGridKind.node, self.topology.node_count)
        raise ValueError("Data array did not have any face, edge, or node dimension")

    @cached_property
    def polygons(self) -> np.ndarray:
        """Generate list of Polygons"""
        # X,Y coords of each node
        topology = self.topology
        node_x = topology.node_x.values
        node_y = topology.node_y.values
        face_node = topology.face_node_array
        polygons = np.full(topology.face_count, None, dtype=np.object_)

        # `shapely.polygons` will make polygons with the same number of vertices.
        # UGRID polygons have arbitrary numbers of vertices.
        # Group polygons by how many vertices they have, then make them in bulk.
        polygons_of_size: Mapping[int, Dict[int, np.ndarray]] = defaultdict(dict)
        for index, row in enumerate(face_node):
            vertices = row.compressed()
            polygons_of_size[vertices.size][index] = np.c_[node_x[vertices], node_y[vertices]]

        for size, size_polygons in polygons_of_size.items():
            coords = np.stack(list(size_polygons.values()))
            shapely.polygons(coords, indices=list(size_polygons.keys()), out=polygons)

        return polygons

    @cached_property
    def face_centres(self) -> np.ndarray:
        face_x, face_y = self.topology.face_x, self.topology.face_y
        if face_x is not None and face_y is not None:
            face_centres = np.column_stack((face_x, face_y))
            return cast(np.ndarray, face_centres)
        return super().face_centres

    def selector_for_index(self, index: UGridIndex) -> Dict[Hashable, int]:
        kind, i = index
        if kind is UGridKind.face:
            return {self.topology.face_dimension: i}
        if kind is UGridKind.edge:
            if self.topology.has_edge_dimension:
                return {self.topology.edge_dimension: i}
            else:
                raise ValueError("Grid has no edge dimension")
        if kind is UGridKind.node:
            return {self.topology.node_dimension: i}
        raise ValueError("Invalid index")  # pragma: no-cover

    def make_linear(self, data_array: xr.DataArray) -> xr.DataArray:
        grid_kind, shape = self.get_grid_kind_and_size(data_array)
        grid_dimension = self.topology.dimension_for_grid_kind[grid_kind]
        return utils.linearise_dimensions(data_array, [grid_dimension])

    def make_clip_mask(
        self,
        clip_geometry: BaseGeometry,
        buffer: int = 0,
    ) -> xr.Dataset:
        """
        Make a mask dataset from a clip geometry for this dataset.
        This mask can later be applied using :meth:`apply_clip_mask`.

        A mask for a UGRID dataset indicates which nodes, edges, and faces
        (collectively 'elements') to include, and their new indices. When
        elements are dropped from the dataset, to keep the remaining data
        contiguous, all elements gain new indices.

        The mask consists of three data arrays corresponding to nodes, edges,
        and faces. Each data array indicates the new index for that element, or
        whether that element should be dropped.

        Making a mask consists of three steps:

        * Identifying which polygons to include
        * Adding a buffer zone around these faces
        * Making a mask for the faces, edges, and nodes
        """
        # Find all faces that intersect the clip geometry
        logger.info("Making clip mask")
        face_indices = np.array([
            item.linear_index
            for polygon, item in self.spatial_index.query(clip_geometry)
            if polygon.intersects(clip_geometry)
        ])
        logger.debug("Found %d intersecting faces, adding size %d buffer...", len(face_indices), buffer)

        # Include all the neighbours of the intersecting faces
        for _ in range(buffer):
            face_indices = buffer_faces(face_indices, self.topology)
        logger.debug("Total faces in mask: %d", len(face_indices))

        # Make a mask dataset
        return mask_from_face_indices(face_indices, self.topology)

    def apply_clip_mask(self, clip_mask: xr.Dataset, work_dir: Pathish) -> xr.Dataset:
        """
        Make a new dataset by applying a clip mask to this dataset.

        See also:
        * :meth:`make_clip_mask`
        * :func:`mask_from_face_indices`
        """
        logger.info("Applying clip mask")
        dataset = self.dataset
        topology = self.topology
        work_path = pathlib.Path(work_dir)

        # Each variable will be sliced and saved to a separate file in the
        # working directory, then reassembled later as a multifile dataset.
        # This list keeps track of all the individual files.
        mfdataset_paths = []

        logger.debug("Slicing topology variables...")
        # Collect all the topology variables here. These need special handling,
        # compared to data variables. The mesh variable can be reused without
        # any changes.
        topology_variables: List[xr.DataArray] = [topology.mesh_variable]

        def integer_indices(data_array: xr.DataArray) -> np.ndarray:
            masked = np.ma.masked_invalid(data_array.values)
            masked_integers: np.ndarray = masked.astype(np.int_)
            return masked_integers

        # This is the fill value used in the mask.
        new_fill_value = clip_mask.data_vars['new_node_index'].encoding['_FillValue']
        new_node_indices = integer_indices(clip_mask.data_vars['new_node_index'])
        new_face_indices = integer_indices(clip_mask.data_vars['new_face_index'])
        has_edges = 'new_edge_index' in clip_mask.data_vars
        if has_edges:
            new_edge_indices = integer_indices(clip_mask.data_vars['new_edge_index'])

        # Re-index the face_node_connectivity variable
        topology_variables.append(update_connectivity(
            topology.face_node_connectivity, topology.face_node_array,
            new_face_indices, new_node_indices,
            primary_dimension=topology.face_dimension, fill_value=new_fill_value))

        # Re-index each of the optional connectivity variables
        if has_edges and topology.has_valid_face_edge_connectivity:
            topology_variables.append(update_connectivity(
                topology.face_edge_connectivity, topology.face_edge_array,
                new_face_indices, new_edge_indices,
                primary_dimension=topology.edge_dimension, fill_value=new_fill_value))

        if topology.has_valid_face_face_connectivity:
            topology_variables.append(update_connectivity(
                topology.face_face_connectivity, topology.face_face_array,
                new_face_indices, new_face_indices,
                primary_dimension=topology.face_dimension, fill_value=new_fill_value))

        if has_edges and topology.has_valid_edge_face_connectivity:
            topology_variables.append(update_connectivity(
                topology.edge_face_connectivity, topology.edge_face_array,
                new_edge_indices, new_face_indices,
                primary_dimension=topology.face_dimension, fill_value=new_fill_value))

        if has_edges and topology.has_valid_edge_node_connectivity:
            topology_variables.append(update_connectivity(
                topology.edge_node_connectivity, topology.edge_node_array,
                new_edge_indices, new_node_indices,
                primary_dimension=topology.edge_dimension, fill_value=new_fill_value))

        # Save all the topology variables to one combined dataset
        topology_path = work_path / (str(topology.mesh_variable.name) + ".nc")
        topology_dataset = xr.Dataset(
            data_vars={variable.name: variable for variable in topology_variables},
            coords=dataset.coords,
        )
        topology_dataset.to_netcdf(topology_path)
        mfdataset_paths.append(topology_path)

        topology_variable_names = list(topology_dataset.variables.keys())
        del topology_dataset
        del topology_variables

        logger.debug("Slicing data variables...")
        dimension_masks: Dict[Hashable, np.ndarray] = {
            topology.node_dimension: ~np.ma.getmask(new_node_indices),
            topology.face_dimension: ~np.ma.getmask(new_face_indices),
        }
        if has_edges:
            dimension_masks[topology.edge_dimension] = ~np.ma.getmask(new_edge_indices)
        mesh_dimensions = set(dimension_masks.keys())

        for name, data_array in dataset.data_vars.items():
            data_array_path = work_path / (str(name) + '.nc')
            if name in topology_variable_names:
                logger.debug("Skipping %r as it is a topology variable", name)

            elif set(data_array.dims).isdisjoint(mesh_dimensions):
                logger.debug("Using %r as-is, as it has no mesh dimensions", name)
                data_array = data_array.copy()
                utils.disable_default_fill_value(data_array)
                data_array.to_netcdf(data_array_path)
                mfdataset_paths.append(data_array_path)
                del data_array

            else:
                logger.debug("Slicing %r, dimensions %r", name, data_array.dims)
                # This is basically doing `data_array.isel({dimension: rows})`,
                # except that `isel` does this _very_ slowly.
                values = data_array.values
                for index, dim in enumerate(data_array.dims):
                    # For each dimension in the data array, if it is one of
                    # the dimensions that we want to subset on...
                    if dim in dimension_masks:
                        # ... make a tuple slice like (:, :, :, bool_array)
                        # where bool_array indicates which rows to include
                        slice_index = tuple([np.s_[:]] * index + [dimension_masks[dim]])  # type: ignore
                        values = values[slice_index]

                data_array = xr.DataArray(data=values, dims=data_array.dims, name=name)
                utils.disable_default_fill_value(data_array)
                data_array.to_netcdf(data_array_path)
                mfdataset_paths.append(data_array_path)

                del data_array
                del values

        logger.debug("Merging individual variables...")
        new_dataset = xr.open_mfdataset(mfdataset_paths, lock=False)
        return utils.dataset_like(dataset, new_dataset)

    def drop_geometry(self) -> xr.Dataset:
        dataset = self.dataset
        topology = self.topology

        geometry_variables = [
            topology.mesh_variable.name,
            topology.face_node_connectivity.name,
            topology.node_x.name,
            topology.node_y.name,
        ]
        if topology.has_valid_face_edge_connectivity:
            geometry_variables.append(topology.face_edge_connectivity.name)
        if topology.has_valid_face_face_connectivity:
            geometry_variables.append(topology.face_face_connectivity.name)
        if topology.has_valid_edge_node_connectivity:
            geometry_variables.append(topology.edge_node_connectivity.name)
        if topology.has_valid_edge_face_connectivity:
            geometry_variables.append(topology.edge_face_connectivity.name)
        if topology.edge_x is not None:
            geometry_variables.append(topology.edge_x.name)
        if topology.edge_y is not None:
            geometry_variables.append(topology.edge_y.name)
        if topology.face_x is not None:
            geometry_variables.append(topology.face_x.name)
        if topology.face_y is not None:
            geometry_variables.append(topology.face_y.name)

        dataset = dataset.drop_vars(geometry_variables)
        dataset.attrs.pop('Conventions', None)

        return dataset



import datetime
import pathlib

import netCDF4
import numpy as np
import numpy.testing
import pandas as pd
import pytest
import xarray
import xarray.testing

from emsarray import utils


@pytest.mark.parametrize(
    ('existing', 'new'),
    [
        ('days since 1990-01-01T00:00:00+10:00', 'days since 1990-01-01 00:00:00 +10:00'),
        ('days since 1990-1-1 0:00:00 +10', 'days since 1990-01-01 00:00:00 +10:00'),
        ('hours since 2021-11-16T12:00:00+11:00', 'hours since 2021-11-16 12:00:00 +11:00'),
    ],
)
def test_format_time_units_for_ems(existing, new):
    assert new == utils.format_time_units_for_ems(existing)


def test_fix_time_units_for_ems(tmp_path: pathlib.Path):
    dataset_path = tmp_path / "dataset.nc"
    # Make a minimal dataset, with a record dimension and time data
    xr_dataset = xarray.Dataset(
        data_vars={
            't': xarray.DataArray(
                data=[datetime.datetime(2021, 11, 16, hour) for hour in range(24)],
                dims=["record"],
                attrs={'coordinate': 'time', 'long_name': 'Time'},
            ),
        },
    )
    # Set a nice, EMS compatible time unit string
    xr_dataset.data_vars['t'].encoding.update({
        'units': 'hours since 2021-11-01 00:00:00 +10:00',
        'calendar': 'proleptic_gregorian',
    })
    xr_dataset.encoding['unlimited_dims'] = {'record'}
    xr_dataset.to_netcdf(dataset_path)

    with netCDF4.Dataset(dataset_path, "r") as nc_dataset:
        # xarray will have reformatted the time units to ISO8601 format,
        # with a 'T' separator and no space before the timezone
        assert nc_dataset.variables['t'].getncattr('units') == 'hours since 2021-11-01T00:00:00+10:00'

    utils.fix_time_units_for_ems(dataset_path, "t")

    with netCDF4.Dataset(dataset_path, "r+") as nc_dataset:
        # The time units should now be in an EMS-compatible format
        assert nc_dataset.variables['t'].getncattr('units') == 'hours since 2021-11-01 00:00:00 +10:00'


def test_disable_default_fill_value(tmp_path: pathlib.Path):
    foo = xarray.DataArray(
        data=np.arange(35, dtype=int).reshape(5, 7),
        dims=['j', 'i'],
        attrs={"Hello": "World"},
    )
    bar = xarray.DataArray(
        data=np.arange(35, dtype=np.float64).reshape(5, 7),
        dims=['j', 'i'],
    )
    baz = xarray.DataArray(
        data=np.arange(35, dtype=np.float64).reshape(5, 7),
        dims=['j', 'i'],
    )
    baz.data = np.where(np.tri(5, 7, dtype=bool), baz.data, np.nan)
    baz.encoding["_FillValue"] = np.nan

    dataset = xarray.Dataset(data_vars={"foo": foo, "bar": bar, "baz": baz})

    # Save to a netCDF4 and then prove that it is bad
    dataset.to_netcdf(tmp_path / "bad.nc")
    with netCDF4.Dataset(tmp_path / "bad.nc", "r") as nc_dataset:
        # This one shouldn't be here because it is an integer datatype. xarray
        # does the right thing already in this case.
        assert '_FillValue' not in nc_dataset.variables["foo"].ncattrs()
        # This one shouldn't be here as we didnt set it, and the array is full!
        # This is the problem we are trying to solve
        assert np.isnan(nc_dataset.variables["bar"].getncattr("_FillValue"))
        # This one is quite alright, we did explicitly set it after all
        assert np.isnan(nc_dataset.variables["baz"].getncattr("_FillValue"))

    utils.disable_default_fill_value(dataset)
    dataset.to_netcdf(tmp_path / "good.nc")
    with netCDF4.Dataset(tmp_path / "good.nc", "r") as nc_dataset:
        # This one should still be unset
        assert '_FillValue' not in nc_dataset.variables["foo"].ncattrs()
        # This one should now be unset
        assert '_FillValue' not in nc_dataset.variables["bar"].ncattrs()
        # Make sure this didn't get clobbered
        assert np.isnan(nc_dataset.variables["baz"].getncattr("_FillValue"))


def test_dataset_like():
    sample_foo = xarray.DataArray(
        data=np.arange(10, dtype=np.int32),
        coords=[np.arange(10) * 2], dims=['x'])
    sample_foo.attrs = {'units': 'meters'}
    sample_foo.encoding = {'_FillValue': 10}
    sample_bar = xarray.DataArray(
        data=np.arange(20, dtype=np.double),
        coords=[list("abcdefghijklmnopqrst")], dims=['y'])
    sample_bar.attrs = {'units': 'seconds', 'long_name': 'Seconds since lunch'}
    sample_bar.encoding = {'_FillValue': np.nan}
    sample_dataset = xarray.Dataset({'foo': sample_foo, 'bar': sample_bar})
    sample_dataset.attrs = {"hello": "world", "from": "sample"}
    sample_dataset.encoding = {"reticulation": "spline"}

    new_foo = xarray.DataArray(
        data=np.arange(5, dtype=np.int32), coords=[np.arange(5) * 3], dims=['x'])
    new_bar = xarray.DataArray(
        data=np.arange(10, dtype=np.double), coords=[list("ABCDEFGHIJ")], dims=['y'])
    new_bar.attrs = {'long_name': 'Seconds since breakfast'}
    new_bar.encoding = {'_FillValue': -9999.}
    new_dataset = xarray.Dataset({'bar': new_bar, 'foo': new_foo})
    new_dataset.attrs = {"from": "new"}
    new_dataset.encoding = {"reticulation": "quartic"}

    # Variables have an order in the dataset. The order in the new dataset is wrong
    assert list(sample_dataset.data_vars.keys()) == ['foo', 'bar']
    assert list(new_dataset.data_vars.keys()) == ['bar', 'foo']
    assert list(sample_dataset.dims.keys()) == ['x', 'y']
    assert list(new_dataset.dims.keys()) == ['y', 'x']

    # Make a new dataset like the input dataset
    like_dataset = utils.dataset_like(sample_dataset, new_dataset)

    # The data variables and dimensions should now be identical
    assert list(like_dataset.data_vars.keys()) == ['foo', 'bar']
    assert list(like_dataset.dims.keys()) == ['x', 'y']
    # The attributes and encodings should be merged
    assert like_dataset.attrs == {"hello": "world", "from": "new"}
    assert like_dataset.encoding == {"reticulation": "quartic"}

    # The data should remain the same.
    # The attributes and encoding should be merged.
    like_foo = like_dataset.data_vars['foo']
    numpy.testing.assert_equal(like_foo.values, new_foo.values)
    assert like_foo.attrs == {'units': 'meters'}
    assert like_foo.encoding['_FillValue'] == 10
    like_bar = like_dataset.data_vars['bar']
    numpy.testing.assert_equal(like_bar.values, new_bar.values)
    assert like_bar.attrs == {'units': 'seconds', 'long_name': 'Seconds since breakfast'}
    assert like_bar.encoding['_FillValue'] == -9999.


def test_extract_vars():
    time_size, depth_size = 4, 10
    lon_size, lat_size = 10, 15
    lon_grid, lat_grid = np.arange(lon_size + 1), np.arange(lat_size + 1)

    dataset = xarray.Dataset(
        data_vars={
            'lon_bounds': (["lon", 2], np.stack([lon_grid[:-1], lon_grid[1:]], axis=-1)),
            'lat_bounds': (["lat", 2], np.stack([lat_grid[:-1], lat_grid[1:]], axis=-1)),
            'botz': (["lat", "lon"], 50 + 25 * np.random.random_sample((lat_size, lon_size))),
            'eta': (
                ["time", "lat", "lon"],
                np.random.random_sample((time_size, lat_size, lon_size))),
            'temp': (
                ["time", "depth", "lat", "lon"],
                15 + 3 * np.random.random_sample((time_size, depth_size, lat_size, lon_size))),
            'salt': (
                ["time", "depth", "lat", "lon"],
                34 + np.random.random_sample((time_size, depth_size, lat_size, lon_size))),
        },
        coords={
            'time': (["time"], pd.date_range("2021-12-21", periods=time_size)),
            'depth': (["depth"], np.arange(depth_size), {"positive": "down"}),
            'lon': (["lon"], (lon_grid[1:] + lon_grid[:-1]) / 2, {"bounds": "lon_bounds"}),
            'lat': (["lat"], (lat_grid[1:] + lat_grid[:-1]) / 2, {"bounds": "lat_bounds"}),
        },
    )

    eta_botz_and_bounds = utils.extract_vars(dataset, ["eta", "botz"])
    # Bounds variables for the coordinates are included
    assert set(eta_botz_and_bounds.data_vars.keys()) == {'lon_bounds', 'lat_bounds', 'eta', 'botz'}
    # Note that 'depth' is included in the coords, even though no variables use it
    assert set(eta_botz_and_bounds.coords.keys()) == {'time', 'depth', 'lon', 'lat'}

    salt_temp = utils.extract_vars(dataset, ["salt", "temp"], keep_bounds=False)
    # No more bounds variables
    assert set(salt_temp.data_vars.keys()) == {'salt', 'temp'}
    # All the coords are still kept
    assert set(salt_temp.coords.keys()) == {'time', 'depth', 'lon', 'lat'}

    with pytest.raises(ValueError):
        utils.extract_vars(dataset, ['eta', 'cloud'])

    just_eta = utils.extract_vars(dataset, ['eta', 'cloud'], keep_bounds=False, errors='ignore')
    assert set(just_eta.data_vars.keys()) == {'eta'}


def test_check_data_array_dimensions_match_complete():
    """
    check_data_array_dimensions_match with all dimensions present and matching
    """
    time_size = 4
    lon_size, lat_size = 10, 15

    dataset = xarray.Dataset(
        data_vars={
            'botz': (["lat", "lon"], 50 + 25 * np.random.random_sample((lat_size, lon_size))),
            'eta': (
                ["time", "lat", "lon"],
                np.random.random_sample((time_size, lat_size, lon_size)),
            ),
        },
        coords={
            'time': (["time"], pd.date_range("2021-12-21", periods=time_size)),
            'lon': (["lon"], np.arange(lon_size) + 0.5),
            'lat': (["lat"], np.arange(lat_size) + 0.5),
        }
    )

    # This data array has the same time/lat/lon dimension size as the dataset
    # above, but is otherwise unrelated. The data can still be plotted using
    # the spatial data of the dataset, assuming the dimensions match.
    surface_temp = xarray.DataArray(
        data=15 + 3 * np.random.random_sample((time_size, lat_size, lon_size)),
        dims=['time', 'lat', 'lon'],
    )

    utils.check_data_array_dimensions_match(dataset, surface_temp)


def test_check_data_array_dimensions_match_subset():
    """
    check_data_array_dimensions_match with a subset of dimensions present and matching
    """
    time_size, depth_size = 4, 3
    lon_size, lat_size = 10, 15

    dataset = xarray.Dataset(
        data_vars={
            'botz': (["lat", "lon"], 50 + 25 * np.random.random_sample((lat_size, lon_size))),
            'temp': (
                ["time", "depth", "lat", "lon"],
                np.random.random_sample((time_size, depth_size, lat_size, lon_size)),
            ),
        },
        coords={
            'time': (["time"], pd.date_range("2021-12-21", periods=time_size)),
            'depth': (["depth"], np.arange(depth_size), {'positive': 'down'}),
            'lon': (["lon"], np.arange(lon_size) + 0.5),
            'lat': (["lat"], np.arange(lat_size) + 0.5),
        }
    )

    # By slicing off one layer and dropping the depth dimension, the data array
    # dimensions should still match as a subset of the dataset dimensions.
    surface_temp = dataset.data_vars['temp'].isel(depth=0, drop=True)

    utils.check_data_array_dimensions_match(dataset, surface_temp)


def test_check_data_array_dimensions_match_size_mismatch():
    """
    check_data_array_dimensions_match with some dimensions having different sizes
    """
    time_size, depth_size = 4, 3
    lon_size, lat_size = 10, 15

    dataset = xarray.Dataset(
        data_vars={
            'botz': (["lat", "lon"], 50 + 25 * np.random.random_sample((lat_size, lon_size))),
            'temp': (
                ["time", "depth", "lat", "lon"],
                np.random.random_sample((time_size, depth_size, lat_size, lon_size)),
            ),
        },
        coords={
            'time': (["time"], pd.date_range("2021-12-21", periods=time_size)),
            'depth': (["depth"], np.arange(depth_size), {'positive': 'down'}),
            'lon': (["lon"], np.arange(lon_size) + 0.5),
            'lat': (["lat"], np.arange(lat_size) + 0.5),
        }
    )

    # Slicing the temperature to get a subset of the lat/lon grid will cause
    # the dimensions to differ
    surface_temp = dataset.data_vars['temp'].isel(
        {'lon': np.s_[2:-2], 'lat': np.s_[3:-3], 'depth': 0},
        drop=True)

    with pytest.raises(ValueError):
        utils.check_data_array_dimensions_match(dataset, surface_temp)

    # If you subset the dataset as well, this should work fine though!
    dataset_subset = dataset.isel({'lon': np.s_[2:-2], 'lat': np.s_[3:-3]})
    surface_temp_subset = dataset_subset.data_vars['temp'].isel({'depth': 0}, drop=True)
    utils.check_data_array_dimensions_match(dataset_subset, surface_temp_subset)


def test_check_data_array_dimensions_match_unknown_dimension():
    """
    check_data_array_dimensions_match with an unknown dimension
    """
    time_size, depth_size = 4, 3
    lon_size, lat_size = 10, 15

    dataset = xarray.Dataset(
        data_vars={
            'botz': (["lat", "lon"], 50 + 25 * np.random.random_sample((lat_size, lon_size))),
            'eta': (
                ["time", "lat", "lon"],
                np.random.random_sample((time_size, lat_size, lon_size)),
            ),
        },
        coords={
            'time': (["time"], pd.date_range("2021-12-21", periods=time_size)),
            'lon': (["lon"], np.arange(lon_size) + 0.5),
            'lat': (["lat"], np.arange(lat_size) + 0.5),
        }
    )

    # Slicing the temperature to get a subset of the lat/lon grid will cause
    # the dimensions to differ
    temp = xarray.DataArray(
        data=15 + 3 * np.random.random_sample((time_size, depth_size, lat_size, lon_size)),
        dims=['time', 'depth', 'lat', 'lon'],
    )

    with pytest.raises(ValueError):
        utils.check_data_array_dimensions_match(dataset, temp)

    # If you remove the extra dimension, this should be fine
    surface_temp = temp.isel({'depth': 0}, drop=True)
    utils.check_data_array_dimensions_match(dataset, surface_temp)


def test_move_dimensions_to_end():
    data_array = xarray.DataArray(
        data=np.random.random((2, 7, 5, 3)),
        dims=['t', 'x', 'y', 'z'],
    )

    # This should result in a dimension order of ('t', 'z', 'y', 'z')
    transposed = utils.move_dimensions_to_end(data_array, ['y', 'x'])

    # Check that the dimensions are in the correct order, with the correct shape
    assert transposed.dims == ('t', 'z', 'y', 'x')
    assert transposed.shape == (2, 3, 5, 7)

    # Check that the values were rearranged as expected
    numpy.testing.assert_equal(
        transposed.isel(t=1, z=2).values,
        np.transpose(data_array.isel(t=1, z=2).values))

    # This should be a no-op, as those dimensions are already at the end
    xarray.testing.assert_equal(
        data_array,
        utils.move_dimensions_to_end(data_array, ['y', 'z']))


def test_move_dimensions_to_end_missing():
    data_array = xarray.DataArray(
        data=np.random.random((2, 7, 5, 3)),
        dims=['t', 'x', 'y', 'z'],
    )

    with pytest.raises(ValueError) as exc:
        utils.move_dimensions_to_end(data_array, ['x', 'foo', 'bar'])

    message = "DataArray does not contain dimensions ['bar', 'foo']"
    assert str(exc.value) == message


def test_linearize_dimensions_exact_dimensions():
    data_array = xarray.DataArray(
        data=np.random.random((3, 5)),
        dims=['y', 'x'],
    )
    expected = xarray.DataArray(
        data=data_array.values.ravel(),
        dims=['index'],
    )
    linearized = utils.linearise_dimensions(data_array, ['y', 'x'])
    xarray.testing.assert_equal(linearized, expected)


def test_linearize_dimensions_extra_dimensions():
    data_array = xarray.DataArray(
        data=np.random.random((2, 7, 5, 3)),
        dims=['t', 'x', 'y', 'z'],
        coords={
            't': pd.date_range("2022-03-02", periods=2),
            'x': np.arange(7),
            'y': np.arange(5),
            'z': [0.25, 0.5, 1.5],
        }
    )
    expected = xarray.DataArray(
        data=np.reshape(np.transpose(data_array.values, (0, 3, 2, 1)), (2, 3, -1)),
        dims=['t', 'z', 'index'],
        coords={
            't': data_array.coords['t'],
            'z': data_array.coords['z'],
        },
    )
    linearized = utils.linearise_dimensions(data_array, ['y', 'x'])
    xarray.testing.assert_equal(linearized, expected)


def test_linearize_dimensions_custom_name():
    data_array = xarray.DataArray(
        data=np.random.random((2, 7, 5, 3)),
        dims=['t', 'x', 'y', 'z'],
    )
    expected = xarray.DataArray(
        data=np.reshape(np.transpose(data_array.values, (0, 3, 2, 1)), (2, 3, -1)),
        dims=['t', 'z', 'i'],
    )
    linearized = utils.linearise_dimensions(data_array, ['y', 'x'], linear_dimension='i')
    xarray.testing.assert_equal(linearized, expected)


def test_linearize_dimensions_auto_name_conflict():
    data_array = xarray.DataArray(
        data=np.random.random((2, 7, 5, 3)),
        dims=['index', 'index_0', 'index_1', 'index_2'],
    )
    expected = xarray.DataArray(
        data=np.reshape(np.transpose(data_array.values, (0, 1, 3, 2)), (2, 7, -1)),
        dims=['index', 'index_0', 'index_1'],
    )
    linearized = utils.linearise_dimensions(data_array, ['index_2', 'index_1'])
    xarray.testing.assert_equal(linearized, expected)



"""
Test binding Convention instances to datasets,
whether implicitly through the accessor autodetection or manually.
"""
import pytest
import xarray as xr

import emsarray
from emsarray.conventions import CFGrid2D, ShocStandard
from emsarray.state import State


def test_automatic_binding_via_accessor(datasets):
    ds = xr.open_dataset(datasets / 'cfgrid1d.nc')
    state = State.get(ds)

    # Fresh datasets opened via xarray should be unbound initially
    assert not state.is_bound()
    assert state.convention is None

    # Autodetection via the accessor should bind the convention
    convention = ds.ems
    assert state.is_bound()
    assert state.convention is convention


def test_automatic_binding_via_open_dataset(datasets):
    ds = emsarray.open_dataset(datasets / 'cfgrid1d.nc')
    state = State.get(ds)

    # Datasets opened via emsarray.open_dataset should be bound
    assert state.is_bound()
    assert ds.ems is state.convention


def test_manual_binding(datasets):
    ds = xr.open_dataset(datasets / 'shoc_standard.nc')
    state = State.get(ds)

    # Fresh datasets opened via xarray should be unbound initially
    assert not state.is_bound()
    assert state.convention is None

    # Construct a convention. Autodetection would use ShocStandard,
    # but CFGrid2D is also compatible.
    convention = CFGrid2D(ds, longitude='x_centre', latitude='y_centre')

    # Merely constructing this convention should not bind anything
    assert not state.is_bound()
    assert state.convention is None

    # Manually bind and check that it holds
    convention.bind()
    assert state.is_bound()
    assert state.convention is convention
    assert ds.ems is convention


def test_copy_rebind(datasets):
    ds_shoc = emsarray.open_dataset(datasets / 'shoc_standard.nc')
    shoc_convention = ds_shoc.ems
    assert isinstance(shoc_convention, ShocStandard)

    # Binding an already bound dataset should raise an error
    cfgrid_convention = CFGrid2D(ds_shoc, longitude='x_centre', latitude='y_centre')
    with pytest.raises(ValueError):
        cfgrid_convention.bind()

    # Making a copy and binding that should work
    ds_cfgrid = ds_shoc.copy()
    cfgrid_convention = CFGrid2D(ds_cfgrid, longitude='x_centre', latitude='y_centre')
    cfgrid_convention.bind()

    # Each dataset should have its own bound convention
    assert ds_cfgrid.ems is cfgrid_convention
    assert ds_shoc.ems is shoc_convention
